import inspect
from functools import wraps


def filter_kwargs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        arg_names = inspect.getfullargspec(func).args
        correct_kwargs = {}
        for arg in arg_names:
            if arg in kwargs:
                correct_kwargs[arg] = kwargs[arg]
        return func(*args, **correct_kwargs)

    return wrapper


def tqdm_parallel_decorator(func):
    def wrapper(progress_bar, *args, **kwargs):
        output = func(*args, **kwargs)
        progress_bar.update(1)
        return output

    return wrapper



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Subpackage for experiment and pilots design.
"""
from .designer import (
    Designer,
    design,
    design_binary,
    design_binary_effect,
    design_binary_power,
    design_binary_size,
    load_from_config,
)

__all__ = [
    "Designer",
    "design",
    "load_from_config",
    "design_binary_size",
    "design_binary_effect",
    "design_binary_power",
    "design_binary",
]



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Tuple

import numpy as np
import scipy.stats as sps

import ambrosia.tools.pvalue_tools as pvalue_pkg
import ambrosia.tools.theoretical_tools as theory_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import ABStatCriterion, StatCriterion


def get_results_dict(alpha: float, pvalue: float, effect: float, conf_int: Tuple[float, float]):
    """
    Returns dictionary with given values.
    """
    return {"first_type_error": alpha, "pvalue": pvalue, "effect": effect, "confidence_interval": conf_int}


def get_calc_effect_ttest(group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute"):
    """
    Calculation effect for ttest.
    """
    if effect_type == "absolute":
        return np.mean(group_b, axis=0) - np.mean(group_a, axis=0)
    elif effect_type == "relative":
        return np.mean(group_b, axis=0) / np.mean(group_a, axis=0) - 1
    else:
        raise ValueError(ABStatCriterion._send_type_error_msg())  # pylint: disable=W0212


class TtestIndCriterion(ABStatCriterion):
    """
    Unit for independent T-test.
    """

    implemented_effect_types: List = ["absolute", "relative"]

    def calculate_pvalue(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute", **kwargs):
        if effect_type == "absolute":
            return sps.ttest_ind(a=group_b, b=group_a, equal_var=False, **kwargs).pvalue
        elif effect_type == "relative":
            _, pvalue = theory_pkg.apply_delta_method(group_a, group_b, "fraction", **kwargs)
            return pvalue
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_effect(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute"):
        return get_calc_effect_ttest(group_a, group_b, effect_type)

    def _build_intervals_absolute(
        self,
        center: float,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        alternative: str = "two-sided",
    ):
        """
        Helps handle different alternatives and dimension for student distribution.
        """
        alpha_corrected: float = pvalue_pkg.corrected_alpha(alpha, alternative)
        quantiles, std_error = theory_pkg.get_ttest_info(group_a, group_b, alpha_corrected)
        left_ci: np.ndarray = center - quantiles * std_error
        right_ci: np.ndarray = center + quantiles * std_error
        left_ci, right_ci = pvalue_pkg.choose_from_bounds(left_ci, right_ci, alternative)
        conf_intervals = list(zip(left_ci, right_ci))
        return conf_intervals

    def calculate_conf_interval(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        effect_type: str = "absolute",
        **kwargs,
    ):
        if isinstance(alpha, float):
            alpha = np.array([alpha])
        if effect_type == "absolute":
            difference_estimation: float = group_b.mean() - group_a.mean()
            conf_intervals = self._build_intervals_absolute(difference_estimation, group_a, group_b, alpha, **kwargs)
        elif effect_type == "relative":
            conf_intervals, _ = theory_pkg.apply_delta_method(group_a, group_b, "fraction", alpha, **kwargs)
        else:
            raise ValueError(self._send_type_error_msg())
        return conf_intervals

    def get_results(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = 0.05,
        effect_type: str = "absolute",
        **kwargs,
    ) -> types.StatCriterionResult:
        """
        Override this method from Base class for handle cases with
        """
        if effect_type == "relative":
            conf_int, pvalue = theory_pkg.apply_delta_method(group_a, group_b, "fraction", alpha, **kwargs)
            effect: float = self.calculate_effect(group_a, group_b, effect_type)
            return get_results_dict(alpha, pvalue, effect, conf_int)
        return super().get_results(group_a, group_b, alpha, effect_type, **kwargs)


class TtestRelCriterion(ABStatCriterion):
    """
    Unit for relative paired T-test.
    """

    implemented_effect_types: List = ["absolute", "relative"]

    def calculate_pvalue(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute", **kwargs):
        if effect_type == "absolute":
            return sps.ttest_rel(a=group_b, b=group_a, **kwargs).pvalue
        elif effect_type == "relative":
            _, pvalue = theory_pkg.apply_delta_method(group_a, group_b, "fraction", dependent=True, **kwargs)
            return pvalue
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_effect(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute"):
        return get_calc_effect_ttest(group_a, group_b, effect_type)

    def _build_intervals_absolute(
        self,
        center: float,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        alternative: str = "two-sided",
    ):
        """
        Helps handle different alternatives and build confidence interval
        for related sampels
        """
        alpha_corrected: float = pvalue_pkg.corrected_alpha(alpha, alternative)
        std_error = np.sqrt(np.var(group_b - group_a, ddof=1) / len(group_a))
        quantiles = sps.t.ppf(1 - alpha_corrected / 2, df=len(group_a) - 1)
        left_ci: float = center - quantiles * std_error
        right_ci: float = center + quantiles * std_error
        left_ci, right_ci = pvalue_pkg.choose_from_bounds(left_ci, right_ci, alternative)
        conf_intervals = list(zip(left_ci, right_ci))
        return conf_intervals

    def calculate_conf_interval(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        effect_type: str = "absolute",
        **kwargs,
    ):
        if isinstance(alpha, float):
            alpha = np.array([alpha])
        if effect_type == "absolute":
            difference_estimation: float = np.mean(group_b - group_a)
            conf_intervals = self._build_intervals_absolute(difference_estimation, group_a, group_b, alpha, **kwargs)
        elif effect_type == "relative":
            conf_intervals, _ = theory_pkg.apply_delta_method(
                group_a, group_b, "fraction", alpha, dependent=True, **kwargs
            )
        else:
            raise ValueError(self._send_type_error_msg())
        return conf_intervals

    def get_results(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = 0.05,
        effect_type: str = "absolute",
        **kwargs,
    ) -> types.StatCriterionResult:
        """
        Override this method from Base class for handle cases with
        """
        if effect_type == "relative":
            conf_int, pvalue = theory_pkg.apply_delta_method(
                group_a, group_b, "fraction", alpha, dependent=True, **kwargs
            )
            effect: float = self.calculate_effect(group_a, group_b, "relative")
            return get_results_dict(alpha, pvalue, effect, conf_int)
        return super().get_results(group_a, group_b, alpha, effect_type, **kwargs)


class MannWhitneyCriterion(ABStatCriterion):
    """
    Unit for Mann-Whitney U test.
    """

    implemented_effect_types: List = ["absolute"]

    def calculate_pvalue(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute", **kwargs):
        if effect_type == "absolute":
            return sps.mannwhitneyu(x=group_b, y=group_a, **kwargs).pvalue
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_effect(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute"):
        if effect_type == "absolute":
            return np.median(group_b, axis=0) - np.median(group_a, axis=0)  # to discuss
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_conf_interval(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        effect_type: str = "absolute",
        **kwargs,
    ):
        if isinstance(alpha, float):
            alpha = np.array([alpha])
        if effect_type == "absolute":
            return [(None, None)] * len(alpha)
        else:
            raise ValueError(self._send_type_error_msg())


class WilcoxonCriterion(ABStatCriterion):
    """
    Unit for Wilcoxon paired test.
    """

    implemented_effect_types: List = ["absolute"]

    def calculate_pvalue(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute", **kwargs):
        if effect_type == "absolute":
            return sps.wilcoxon(x=group_b, y=group_a, **kwargs).pvalue
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_effect(self, group_a: np.ndarray, group_b: np.ndarray, effect_type: str = "absolute"):
        if effect_type == "absolute":
            return np.median(group_b, axis=0) - np.median(group_a, axis=0)
        else:
            raise ValueError(self._send_type_error_msg())

    def calculate_conf_interval(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = np.array([0.05]),
        effect_type: str = "absolute",
        **kwargs,
    ):
        if effect_type == "absolute":
            return [(None, None)] * len(alpha)
        else:
            raise ValueError(self._send_type_error_msg())


class ShapiroCriterion(StatCriterion):
    """
    Unit for Shapiro-Wilk test.
    """

    pass



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Tuple

import numpy as np
import scipy.stats as sps

import ambrosia.tools.pvalue_tools as pvalue_pkg
import ambrosia.tools.theoretical_tools as theory_pkg
from ambrosia import types
from ambrosia.spark_tools.theory import get_stats_from_table
from ambrosia.tools.ab_abstract_component import ABStatCriterion
from ambrosia.tools.configs import Effects
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as F
    from pyspark.sql.functions import col, row_number
    from pyspark.sql.window import Window


class ABSparkCriterion(ABStatCriterion):
    """
    Abstract class for Spark criteria
    """

    def _init_cache(self) -> None:
        self.parameters_are_cached: bool = False
        self.data_stats = dict()

    def __init__(self, cache_parameters: bool = True) -> None:
        self.cache_parameters = cache_parameters
        self._init_cache()

    def _delete_cached_data_parameters(self) -> None:
        self.parameters_are_cached = False

    def _calc_and_cache_data_parameters(self, *args, **kwargs) -> None:
        """
        Uses for recalc parameters for cache
        """
        pass

    def _recalc_cache(self, *args, **kwargs) -> None:
        if not self.parameters_are_cached:
            self._calc_and_cache_data_parameters(*args, **kwargs)

    def _check_clear_cache(self) -> None:
        if not self.cache_parameters:
            self._delete_cached_data_parameters()

    def _check_effect(self, effect_type: str) -> None:
        Effects.raise_if_value_incorrect_enum(effect_type)

    def get_results(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        column: str,
        alpha: types.StatErrorType = 0.05,
        effect_type: str = "absolute",
        **kwargs,
    ) -> types.StatCriterionResult:
        return {
            "first_type_error": alpha,
            "pvalue": self.calculate_pvalue(group_a, group_b, column=column, effect_type=effect_type, **kwargs),
            "effect": self.calculate_effect(group_a, group_b, column=column, effect_type=effect_type),
            "confidence_interval": self.calculate_conf_interval(
                group_a, group_b, column=column, alpha=alpha, effect_type=effect_type, **kwargs
            ),
        }


class TtestIndCriterionSpark(ABSparkCriterion):
    """
    Unit for pyspark independent T-test.
    """

    __implemented_effect_types: List = ["absolute", "relative"]
    __type_error_msg: str = f"Choose effect type from {__implemented_effect_types}"
    __data_parameters = ["mean_group_a", "mean_group_b", "std_group_a", "std_group_b", "nobs_group_a", "nobs_group_b"]

    def __calc_and_cache_data_parameters(
        self, group_a: types.SparkDataFrame, group_b: types.SparkDataFrame, column: types.ColumnNameType
    ):
        self.data_stats["mean_group_a"], self.data_stats["std_group_a"] = get_stats_from_table(group_a, column)
        self.data_stats["mean_group_b"], self.data_stats["std_group_b"] = get_stats_from_table(group_b, column)
        self.data_stats["nobs_group_a"] = group_a.count()
        self.data_stats["nobs_group_b"] = group_b.count()
        self.parameters_are_cached = True

    def _apply_delta_method(
        self, alpha: types.StatErrorType = (0.05,), **kwargs
    ) -> Tuple[types.ManyIntervalType, float]:
        if not self.parameters_are_cached:
            raise RuntimeError("Incorrect usage, firstly calculate parameters")
        # Transforms std for delta method
        s1 = theory_pkg.unbiased_to_sufficient(self.data_stats["std_group_a"], self.data_stats["nobs_group_a"])
        s2 = theory_pkg.unbiased_to_sufficient(self.data_stats["std_group_b"], self.data_stats["nobs_group_b"])
        return theory_pkg.apply_delta_method_by_stats(
            size=(self.data_stats["nobs_group_a"] + self.data_stats["nobs_group_a"]) // 2,
            mean_group_a=self.data_stats["mean_group_a"],
            mean_group_b=self.data_stats["mean_group_b"],
            var_group_a=s1**2,
            var_group_b=s2**2,
            alpha=np.array(alpha),
            **kwargs,
        )

    def calculate_pvalue(
        self,
        group_a: types.SparkDataFrame,
        group_b: types.SparkDataFrame,
        column: types.ColumnNameType,
        effect_type: str = "absolute",
        **kwargs,
    ):
        if effect_type not in TtestIndCriterionSpark.__implemented_effect_types:
            raise ValueError(TtestIndCriterionSpark.__type_error_msg)
        if not self.parameters_are_cached:
            self.__calc_and_cache_data_parameters(group_a, group_b, column)
        if effect_type == "absolute":
            p_value = sps.ttest_ind_from_stats(
                self.data_stats["mean_group_b"],
                self.data_stats["std_group_b"],
                self.data_stats["nobs_group_b"],
                self.data_stats["mean_group_a"],
                self.data_stats["std_group_a"],
                self.data_stats["nobs_group_a"],
                **kwargs,
            ).pvalue
        elif effect_type == "relative":
            p_value = self._apply_delta_method(**kwargs)[1]
        if not self.cache_parameters:
            self._delete_cached_data_parameters()
        return p_value

    def calculate_effect(
        self,
        group_a: types.SparkDataFrame,
        group_b: types.SparkDataFrame,
        column: types.ColumnNameType,
        effect_type: str = "absolute",
    ):
        if self.parameters_are_cached is not True:
            self.__calc_and_cache_data_parameters(group_a, group_b, column)
        if effect_type == "absolute":
            effect = self.data_stats["mean_group_b"] - self.data_stats["mean_group_a"]
        elif effect_type == "relative":
            effect = (self.data_stats["mean_group_b"] - self.data_stats["mean_group_a"]) / self.data_stats[
                "mean_group_a"
            ]
        else:
            raise ValueError(TtestIndCriterionSpark.__type_error_msg)
        return effect

    def calculate_conf_interval(
        self,
        group_a: types.SparkDataFrame,
        group_b: types.SparkDataFrame,
        column: types.ColumnNameType,
        alpha: types.StatErrorType = (0.05,),
        effect_type: str = "absolute",
        **kwargs,
    ):
        if self.parameters_are_cached is not True:
            self.__calc_and_cache_data_parameters(group_a, group_b, column)
        if effect_type == "absolute":
            alpha_corrected: float = pvalue_pkg.corrected_alpha(alpha, kwargs["alternative"])
            quantiles, sd = theory_pkg.get_ttest_info_from_stats(
                var_a=self.data_stats["std_group_a"] ** 2,
                var_b=self.data_stats["std_group_b"] ** 2,
                n_obs_a=self.data_stats["nobs_group_a"],
                n_obs_b=self.data_stats["nobs_group_b"],
                alpha=alpha_corrected,
            )
            mean = self.data_stats["mean_group_b"] - self.data_stats["mean_group_a"]
            left_ci: np.ndarray = mean - quantiles * sd
            right_ci: np.ndarray = mean + quantiles * sd
            return self._make_ci(left_ci, right_ci, kwargs["alternative"])
        elif effect_type == "relative":
            conf_interval = self._apply_delta_method(alpha, **kwargs)[0]
            return conf_interval
        else:
            raise ValueError(TtestIndCriterionSpark.__type_error_msg)


class TtestRelativeCriterionSpark(ABSparkCriterion):
    """
    Relative ttest for spark
    """

    __add_index_name: str = "__ambrosia_ind"
    __diff: str = "__ambrosia_rel_diff"
    __ord_col: str = "__ambrosia_ord"

    @staticmethod
    def _rename_col(column: str, group: str) -> str:
        return f"__{column}_{group}"

    def _calc_and_cache_data_parameters(
        self, group_a: types.SparkDataFrame, group_b: types.SparkDataFrame, column: types.ColumnNameType
    ) -> None:
        a_ = (
            group_a.withColumn(self.__ord_col, F.lit(1))
            .withColumn(self.__add_index_name, row_number().over(Window().orderBy(self.__ord_col)))
            .withColumnRenamed(column, self._rename_col(column, "a"))
        )
        b_ = (
            group_b.withColumn(self.__ord_col, F.lit(1))
            .withColumn(self.__add_index_name, row_number().over(Window().orderBy(self.__ord_col)))
            .withColumnRenamed(column, self._rename_col(column, "b"))
        )

        n_a_obs: int = group_a.count()
        n_b_obs: int = group_b.count()

        if n_a_obs != n_b_obs:
            raise ValueError("Size of group A and B must be equal")

        both = a_.join(b_, self.__add_index_name, "inner").withColumn(
            self.__diff, col(self._rename_col(column, "b")) - col(self._rename_col(column, "a"))
        )
        self.data_stats["mean"], self.data_stats["std"] = get_stats_from_table(both, self.__diff)
        self.data_stats["n_obs"] = n_a_obs
        self.parameters_are_cached = True

    def calculate_pvalue(
        self,
        group_a: types.SparkDataFrame,
        group_b: types.SparkDataFrame,
        column: types.ColumnNameType,
        effect_type: str = Effects.abs.value,
        **kwargs,
    ):
        self._recalc_cache(group_a, group_b, column)
        if effect_type == Effects.abs.value:
            p_value = theory_pkg.ttest_1samp_from_stats(
                mean=self.data_stats["mean"], std=self.data_stats["std"], n_obs=self.data_stats["n_obs"], **kwargs
            )
        elif effect_type == Effects.rel.value:
            raise NotImplementedError("Will be implemented later")
        self._check_clear_cache()
        return p_value

    def calculate_conf_interval(
        self,
        group_a: types.SparkDataFrame,
        group_b: types.SparkDataFrame,
        alpha: types.StatErrorType,
        effect_type: str,
        **kwargs,
    ) -> List[Tuple]:
        raise NotImplementedError("Will be implemented later")

    def calculate_effect(
        self, group_a: types.SparkDataFrame, group_b: types.SparkDataFrame, column: str, effect_type: str
    ) -> float:
        self._recalc_cache(group_a, group_b, column)
        if effect_type == Effects.abs.value:
            effect: float = self.data_stats["mean"]
        else:
            raise NotImplementedError("Will be implemented later")
        return effect



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Experiment design methods.

Module contains `Designer` core class and `design` method which are
intended to conduct the experiment design for A/B/.. tests via different
methods.

Experiment design of the individual metric is based on its historical data
and could be done for any parameter from the self-dependent triplet:
group size, effect size and experiment power.

Currently, experiment design problem could be solved using data provided
in form of both pandas and Spark(with some restrictions) dataframes.
"""
from __future__ import annotations

from typing import List, Optional

import numpy as np
import pandas as pd
import yaml

import ambrosia.tools.bin_intervals as bin_pkg
import ambrosia.tools.theoretical_tools as theory_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import ABMetaClass, ABToolAbstract, SimpleDesigner

from .handlers import EmpiricHandler, TheoryHandler, calc_prob_control_class

SIZE: str = "size"
EFFECT: str = "effect"
POWER: str = "power"
BINARY_DESIGN_METHODS: List[str] = ["theory", "binary"]


class Designer(yaml.YAMLObject, ABToolAbstract, metaclass=ABMetaClass):
    """
    Unit for experiments and pilots parameters design.

    Enables to design missing experiment parameters using historical data.
    The main related to each other designable parameters for a single metric are:

        - Effect (Minimal Detectible Effect):
            old_mean_metric_value * effect_value = new_mean_metric_value
        - Sample size:
            Number of research objects in sample
            (for example number of users and their retention).
        - Errors (I type error, II type error):
            I error (alpha):
                Probability to detect presence of effect
                for equally distributed samples.
            II error (beta):
                Probability not to find effect
                for differently distributed samples.

    Parameters
    ----------
    dataframe : PassedDataType, optional
        DataFrame with metrics historical values.
    sizes : SampleSizeType, optional
        Values of research objects number in groups samples during
        the experiment.
    effects : EffectType, optional
        Effects values that are expected during the experiment.
    first_type_errors : StatErrorType, default: ``0.05``
        I type error bounds
        P (detect difference for equal) < alpha.
    second_type_errors : StatErrorType, default: ``0.2``
        II type error bounds
        P (suppose equality for different groups) < beta.
    metrics : MetricNamesType, optional
        Column names of metrics in dataframe to be designed.
    method : str, optional
        Method used for experiment design.
        Can be ``"theory"``, ``"empiric"`` or ``"binary"``.


    Attributes
    ----------
    dataframe : PassedDataType
        DataFrame with metrics historocal values.
    sizes : SampleSizeType
        Number of research objects in group samples.
    effects : EffectType
        Effects values in the experiment.
    first_type_errors : StatErrorType, default: ``0.05``
        I type errors.
    second_type_errors : StatErrorType, default: ``0.2``
        II type errors.
    metrics : MetricNamesType
        Column names of metrics in dataframe to be designed.
    method : str
        Method used for experiment design.

    Examples
    --------
    We have retention labels for users of mobile app for previous month.
    Suppose old_retention = ``0.3``, that is 30% of users returned to the app
    in a month after installation.

    Let us fix the following parameters:
        I type error (alpha) = ``0.05``
        (5% of equal samples we can suppose to be different).

        II type error (beta) = ``0.2``
        (20% of different sampels we can suppose to be equal).

    We add onboarding to our app and want to estimate an effect, by A/B testing
    and wish to increase retention value to 31% percents, so our effect
    parameter gets value of ``1.0(3)``. Now we want to find how much users we
    need in both groups to detect such effect.

    We can use ``Designer`` class in the following way:

    >>> designer = Designer(dataframe=df, metric='retention', effect=1.033)
    >>> designer.run("size")

    Note, that default values for errors are:
        ``first_type_error`` = ``0.05``

        ``second_type_error`` = ``0.2``

    Then we get dataframe that contains value of  sufficient number of users
    for our experiment.

    Notes
    -----

    Constructors:

    >>> designer = Designer()
    >>> # You can pass an Iterable or single object for some parameters
    >>> designer = Designer(
    >>>     dataframe=df,
    >>>     sizes=[100, 200],
    >>>     metrics='LTV',
    >>>     effects=1.05
    >>> )
    >>> designer = Desginer(sizes=1000, metrics=['retention', 'LTV'])
    >>> # You can use path to .csv table for pandas
    >>> designer = Designer('./data/table.csv')

    Setters:

    >>> designer.set_first_errors([0.05, 0.01])
    >>> desginer.set_dataframe(df)

    Run:

    >>> # One can pass arguments and they will have higher priority
    >>> designer.run('size', effects=1.1)
    >>> designer.run('effect', sizes=[500, 1000], metrics='retention')
    >>> # You can set method (watch below)
    >>> designer.run('effect', sizes=[500, 1000], metrics='retention', method='binary')

    Load from yaml config:

    >>> config = '''
            !splitter # <--- this is yaml tag (!important)
                effects:
                    - 0.9
                    - 1.05
                sizes:
                    - 1000
        '''
    >>> designer = yaml.load(config)
    >>> # Or use the implmented function
    >>> designer = load_from_config(config)

    Use standalone function instead of a class:

    >>> design('size', dataframe=df, effects=1.05, metrics='retention')
    """

    # YAML tag for loading from configs
    yaml_tag = "!designer"

    def set_first_errors(self, first_type_errors: types.StatErrorType) -> None:
        if isinstance(first_type_errors, float):
            self.__alpha = [first_type_errors]
        else:
            self.__alpha = first_type_errors

    def set_second_errors(self, second_type_errors: types.StatErrorType) -> None:
        if isinstance(second_type_errors, float):
            self.__beta = [second_type_errors]
        else:
            self.__beta = second_type_errors

    def set_sizes(self, sizes: types.SampleSizeType) -> None:
        if isinstance(sizes, int):
            self.__size = [sizes]
        else:
            self.__size = sizes

    def set_effects(self, effects: types.EffectType) -> None:
        if isinstance(effects, (float, int)):
            self.__effect = [effects]
        else:
            self.__effect = effects

    def set_dataframe(self, dataframe: types.PassedDataType) -> None:
        if isinstance(dataframe, str):
            if dataframe.endswith(".csv"):
                self.__df = pd.read_csv(dataframe)
            else:
                raise ValueError("File name must ends with .csv")
        else:
            self.__df = dataframe

    def set_method(self, method: str) -> None:
        self.__method = method

    def set_metrics(self, metrics: str) -> None:
        if isinstance(metrics, types.MetricNameType):
            self.__metrics = [metrics]
        else:
            self.__metrics = metrics

    def __init__(
        self,
        dataframe: Optional[types.PassedDataType] = None,
        sizes: Optional[types.SampleSizeType] = None,
        effects: Optional[types.EffectType] = None,
        first_type_errors: types.StatErrorType = 0.05,
        second_type_errors: types.StatErrorType = 0.2,
        metrics: Optional[types.MetricNamesType] = None,
        method: str = "theory",
    ):
        """
        Designer class constructor to initialize the object.
        """
        self.set_first_errors(first_type_errors)
        self.set_second_errors(second_type_errors)
        self.set_sizes(sizes)
        self.set_effects(effects)
        self.set_metrics(metrics)
        self.set_dataframe(dataframe)
        self.set_method(method)

    def __getstate__(self):
        """
        Get the state of the object to serialize.
        """
        return dict(
            effects=self.__effect,
            sizes=self.__size,
            first_type_errors=self.__alpha,
            second_type_errors=self.__beta,
            metrics=self.__metrics,
            method=self.__method,
        )

    @classmethod
    def from_yaml(cls, loader: yaml.Loader, node: yaml.Node):
        kwargs = loader.construct_mapping(node)
        return cls(**kwargs)

    @staticmethod
    def __dataframe_handler(handler: SimpleDesigner, parameter: str, **kwargs) -> pd.DataFrame:
        """
        Handles different dataframe types.
        Now pandas and spark are available.
        """
        if parameter == SIZE:
            return handler.size_design(**kwargs)
        elif parameter == EFFECT:
            return handler.effect_design(**kwargs)
        elif parameter == POWER:
            return handler.power_design(**kwargs)
        else:
            raise ValueError(f"Only {SIZE}, {EFFECT} and {POWER} parameters of the experiment could be designed.")

    @staticmethod
    def __theory_design(label: str, args: types._UsageArgumentsType, **kwargs) -> types.DesignerResult:
        """
        Designing an experiment, using a theoretical approach.
        """
        result: types.DesignerResult = {}
        for metric_name in args["metric"]:
            kwargs["dataframe"] = args["df"]
            kwargs["column"] = metric_name
            kwargs["first_errors"] = np.array(args["alpha"])
            if label == SIZE:
                kwargs["effects"] = args[EFFECT]
                kwargs["second_errors"] = np.array(args["beta"])
            elif label == EFFECT:
                kwargs["sample_sizes"] = args[SIZE]
                kwargs["second_errors"] = np.array(args["beta"])
            elif label == POWER:
                kwargs["sample_sizes"] = args[SIZE]
                kwargs["effects"] = args[EFFECT]
            result[metric_name] = Designer.__dataframe_handler(TheoryHandler(), label, **kwargs)
        if len(args["metric"]) == 1:
            return result[args["metric"][0]]
        else:
            return result

    @staticmethod
    def __empiric_design(label: str, args: types._UsageArgumentsType, **kwargs) -> types.DesignerResult:
        """
        Designing an experiment, using an empirical approach.
        """
        kwargs["dataframe"] = args["df"]
        kwargs["alphas"] = np.array(args["alpha"])
        kwargs["metrics"] = args["metric"]
        if label == SIZE:
            kwargs["effects"] = args[EFFECT]
            kwargs["betas"] = np.array(args["beta"])
        elif label == EFFECT:
            kwargs["group_sizes"] = args[SIZE]
            kwargs["betas"] = np.array(args["beta"])
        elif label == POWER:
            groups_ratio: float = kwargs.pop("groups_ratio") if "groups_ratio" in kwargs else 1.0
            kwargs["sample_sizes_a"] = args[SIZE]
            kwargs["sample_sizes_b"] = [int(groups_ratio * size) for size in args[SIZE]]
            kwargs["effects"] = args[EFFECT]
        return Designer.__dataframe_handler(EmpiricHandler(), label, **kwargs)

    @staticmethod
    def __binary_design(label: str, args: types._UsageArgumentsType, **kwargs) -> types.DesignerResult:
        """
        Designing an experiment, using the approach for binary metrics.
        """
        result: types.DesignerResult = {}
        kwargs["first_errors"] = np.array(args["alpha"])
        for metric_name in args["metric"]:
            kwargs["p_a"] = calc_prob_control_class(args["df"], metric_name)
            if label == SIZE:
                kwargs["delta_relative_values"] = args[EFFECT]
                kwargs["second_errors"] = args["beta"]
                result[metric_name] = bin_pkg.get_table_sample_size_on_effect(**kwargs)
            elif label == EFFECT:
                kwargs["second_errors"] = args["beta"]
                kwargs["sample_sizes"] = args[SIZE]
                result[metric_name] = bin_pkg.get_table_effect_on_sample_size(**kwargs)
            elif label == POWER:
                kwargs["delta_relative_values"] = args[EFFECT]
                kwargs["sample_sizes"] = args[SIZE]
                result[metric_name] = bin_pkg.get_table_power_on_size_and_delta(**kwargs)
        if len(args["metric"]) == 1:
            return result[args["metric"][0]]
        else:
            return result

    @staticmethod
    def __pre_design(label: str, args: types._UsageArgumentsType, **kwargs) -> types.DesignerResult:
        """
        Helper function for run() method logic.
        """
        admissible_methods: List[str] = ["theory", "empiric", "binary"]
        if args["method"] == "theory":
            return Designer.__theory_design(label, args, **kwargs)
        elif args["method"] == "empiric":
            return Designer.__empiric_design(label, args, **kwargs)
        elif args["method"] == "binary":
            return Designer.__binary_design(label, args, **kwargs)
        else:
            raise ValueError(f'Choose method from {", ".join(admissible_methods)}, got {args["method"]}')

    def run(
        self,
        to_design: str,
        method: Optional[str] = None,
        sizes: Optional[types.SampleSizeType] = None,
        effects: Optional[types.EffectType] = None,
        first_type_errors: Optional[types.StatErrorType] = None,
        second_type_errors: Optional[types.StatErrorType] = None,
        dataframe: Optional[types.PassedDataType] = None,
        metrics: Optional[types.MetricNamesType] = None,
        **kwargs,
    ) -> types.DesignerResult:
        """
        Perform an experiment design for chosen parameter and metrics
        using historical data.

        Parameters
        ----------
        to_design : str
           Parameter that will be designed using historical data.
           Can take the values of ``"size"``, ``"effect"`` or ``"power"``.
        method : str, optional
            Method used for experiment design.
            Can be ``"theory"``, ``"empiric"`` or ``"binary"``.
        sizes : SampleSizeType, optional
            Values of research objects number in groups samples during
            the experiment.
            If is not provided, must exist as proper class attribute.
        effects : EffectType, optional
            Effects for experiment
            If is not provided, must exist as proper class attribute.
        first_type_errors : StatErrorType, optional
            I type error bounds
            P (detect difference for equal) < alpha.
        second_type_errors : StatErrorType, optional
            II type error bounds
            P (suppose equality for different groups) < beta.
        dataframe : PassedDataType, optional
            DataFrame with metrics historical values.
            If is not provided, must exist as proper class attribute.
        metrics : MetricNamesType, optional
            Column names of metrics in dataframe to be designed.
            If not provided, must exist as proper class attribute.
        **kwargs : Dict
            Other keyword arguments.

        Other Parameters
        ----------------
        as_numeric : bool, default: ``False``
            The result of calculations can be obtained as a percentage string
            either as a number, this parameter could used to toggle.
        groups_ratio : float, default: ``1.0``
            Ratio between two groups.
        alternative : str, default: ``"two-sided"``
            Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
            or ``"less"``.
            ``"greater"`` - if effect is positive.
            ``"less"`` - if effect is negative.
        stabilizing_method : str, default: ``"asin"``
            Effect trasformation. Can be ``"asin"`` and ``"norm"``.
            For non-binary metrics: only ``"norm"`` is accceptable.
            For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
            is more robust and accurate.
            Acceptable only for ``"theory"`` method and actual for binary metrics!

        Returns
        -------
        result : DesignerResult
            Table or dictionary with the results of parameter design for each
            metric.
        """
        if isinstance(effects, (float, int)):
            effects = [effects]
        if isinstance(sizes, int):
            sizes = [sizes]
        if isinstance(first_type_errors, float):
            first_type_errors = [first_type_errors]
        if isinstance(second_type_errors, float):
            second_type_errors = [second_type_errors]
        if isinstance(metrics, types.MetricNameType):
            metrics = [metrics]

        arguments_choice: types._PrepareArgumentsType = {
            "df": (self.__df, dataframe),
            "alpha": (self.__alpha, first_type_errors),
            "metric": (self.__metrics, metrics),
            "method": (self.__method, method),
        }

        designable_parameters: List[str] = [SIZE, EFFECT, POWER]
        if to_design == SIZE:
            arguments_choice[EFFECT] = (self.__effect, effects)
            arguments_choice["beta"] = (self.__beta, second_type_errors)
            chosen_args: types._UsageArgumentsType = Designer._prepare_arguments(arguments_choice)
            return Designer.__pre_design(SIZE, chosen_args, **kwargs)
        elif to_design == EFFECT:
            arguments_choice[SIZE] = (self.__size, sizes)
            arguments_choice["beta"] = (self.__beta, second_type_errors)
            chosen_args: types._UsageArgumentsType = Designer._prepare_arguments(arguments_choice)
            return Designer.__pre_design(EFFECT, chosen_args, **kwargs)
        elif to_design == POWER:
            arguments_choice[SIZE] = (self.__size, sizes)
            arguments_choice[EFFECT] = (self.__effect, effects)
            chosen_args: types._UsageArgumentsType = Designer._prepare_arguments(arguments_choice)
            return Designer.__pre_design(POWER, chosen_args, **kwargs)
        else:
            raise ValueError(f'Incorrect parameter name to design, choose from {", ".join(designable_parameters)}')


def load_from_config(yaml_config: str, loader: type = yaml.Loader) -> Designer:
    """
    Restore a ``Designer`` class instance from a yaml config.

    For yaml_config you can pass file name with config,
    it must ends with .yaml, for example: "config.yaml".

    For loader you can choose SafeLoader.
    """
    if isinstance(yaml_config, str) and yaml_config.endswith(".yaml"):
        with open(yaml_config, "r", encoding="utf8") as file:
            return yaml.load(file, Loader=loader)
    return yaml.load(yaml_config, Loader=loader)


def design(
    to_design,
    dataframe: types.PassedDataType,
    metrics: types.MetricNamesType,
    sizes: types.SampleSizeType = None,
    effects: types.EffectType = None,
    first_type_errors: types.StatErrorType = (0.05,),
    second_type_errors: types.StatErrorType = (0.2,),
    method: str = "theory",
    **kwargs,
) -> types.DesignerResult:
    """
    Function wrapper around the ``Designer`` class.

    Make experiment design based on historical data using passed arguments.

    Creates an instance of the ``Designer`` class internally and execute
    run method with corresponding arguments.

    Parameters
    ----------
    to_design : str
        Parameter that will be designed using historical data.
        Can take the values of ``"size"``, ``"effect"`` or ``"power"``.
    dataframe : PassedDataType
        DataFrame with metrics historical values.
    metrics : MetricNamesType
        Column names of metrics in dataframe to be designed.
    sizes : SampleSizeType, optional
        Values of research objects number in groups samples during
        the experiment.
        If is not provided, ``effects`` value must be defined.
    effects : EffectType, optional
        Effects for experiment
        If is not provided, ``sizes`` value must be defined.
    first_type_errors : StatErrorType, default: ``(0.05,)``
        I type error bounds
        P (detect difference for equal) < alpha.
    second_type_errors : StatErrorType, default: ``(0.2,)``
        II type error bounds
        P (suppose equality for different groups) < beta.
    method : str, default: ``"theory"``
        Method used for experiment design.
        Can be ``"theory"``, ``"empiric"`` or ``"binary"``.
    **kwargs : Dict
        Other keyword arguments.

    Other Parameters
    ----------------
    as_numeric : bool, default: ``False``
        The result of calculations can be obtained as a percentage string
        either as a number, this parameter could used to toggle.
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.
        Acceptable only for ``"theory"`` method and actual for binary metrics!

    Returns
    -------
    result : DesignerResult
        Table or dictionary with the results of parameter design for each
        metric.
    """
    return Designer(
        dataframe=dataframe,
        metrics=metrics,
        first_type_errors=first_type_errors,
        second_type_errors=second_type_errors,
        sizes=sizes,
        effects=effects,
        method=method,
    ).run(to_design, **kwargs)


def design_binary_size(
    prob_a: float,
    effects: types.EffectType,
    first_type_errors: types.StatErrorType = (0.05,),
    second_type_errors: types.StatErrorType = (0.2,),
    method: str = "theory",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
    **kwargs,
) -> pd.DataFrame:
    """
    Design size for binary metrics.

    Parameters
    ----------
    prob_a : float
        Probability of success for the control group.
    effects : EffectType
        List or single value of relative effects.
        For example: ``1.05``, ``[1.05, 1.2]``.
    first_type_errors : StatErrorType, default: ``(0.05,)``
       I type error bounds
       P (detect difference for equal) < alpha.
    second_type_errors : StatErrorType, default: ``(0.2,)``
       II type error bounds
       P (suppose equality for different groups) < beta.
    method : str, default: ``"theory"``
        Supports 2 methods: ``"theory"`` and ``"binary"``
        ``"theory"`` ~ by formula using statsmodels solve_power mechanism
        ``"binary"`` ~ using different types of intervals
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_table : pd.DataFrame
        Table with results of design.
    """
    if isinstance(effects, (float, int)):
        effects = [effects]
    if isinstance(first_type_errors, float):
        first_type_errors = [first_type_errors]
    if isinstance(second_type_errors, float):
        second_type_errors = [second_type_errors]
    if method == "theory":
        return theory_pkg.get_table_sample_size(
            mean=prob_a,
            std=None,
            effects=effects,
            first_errors=first_type_errors,
            second_errors=second_type_errors,
            target_type="binary",
            groups_ratio=groups_ratio,
            alternative=alternative,
            stabilizing_method=stabilizing_method,
        )
    elif method == "binary":
        return bin_pkg.get_table_sample_size_on_effect(
            p_a=prob_a,
            first_errors=first_type_errors,
            second_errors=second_type_errors,
            delta_relative_values=effects,
            **kwargs,
        )
    else:
        raise ValueError(f"Choose valid method from {BINARY_DESIGN_METHODS}, got {method}")


def design_binary_effect(
    prob_a: float,
    sizes: types.SampleSizeType,
    first_type_errors: types.StatErrorType = (0.05,),
    second_type_errors: types.StatErrorType = (0.2,),
    method: str = "theory",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
    as_numeric: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """
    Design effect for binary metrics.

    Parameters
    ----------
    prob_a : float
         Probability of success for the control group.
    sizes : SampleSizeType
        List or single value of group sizes.
        For example: ``100``, ``[100, 200]``.
    first_type_errors : StatErrorType, default: ``(0.05,)``
       I type error bounds
       P (detect difference for equal) < alpha.
    second_type_errors : StatErrorType, default: ``(0.2,)``
       II type error bounds
       P (suppose equality for different groups) < beta.
    method: str, default: ``"theory"``
        Supports 2 methods: ``"theory"`` and ``"binary"``
        ``"theory"`` ~ by formula using statsmodels solve_power mechanism
        ``"binary"`` ~ using different types of intervals
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.
    as_numeric : bool, default: ``False``
        The result of calculations can be obtained as a percentage string
        either as a number, this parameter could used to toggle.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_table : pd.DataFrame
        Table with results of design.
    """
    if isinstance(sizes, int):
        sizes = [sizes]
    if isinstance(first_type_errors, float):
        first_type_errors = [first_type_errors]
    if isinstance(second_type_errors, float):
        second_type_errors = [second_type_errors]
    if method == "theory":
        return theory_pkg.get_minimal_effects_table(
            mean=prob_a,
            std=None,
            sample_sizes=sizes,
            first_errors=first_type_errors,
            second_errors=second_type_errors,
            as_numeric=as_numeric,
            target_type="binary",
            groups_ratio=groups_ratio,
            alternative=alternative,
            stabilizing_method=stabilizing_method,
        )
    elif method == "binary":
        return bin_pkg.get_table_effect_on_sample_size(
            p_a=prob_a,
            sample_sizes=sizes,
            first_errors=first_type_errors,
            second_errors=second_type_errors,
            as_numeric=as_numeric,
            **kwargs,
        )
    else:
        raise ValueError(f"Choose valid method from {BINARY_DESIGN_METHODS}, got {method}")


def design_binary_power(
    prob_a: float,
    sizes: types.SampleSizeType,
    effects: types.EffectType,
    first_type_errors: types.StatErrorType = (0.05,),
    method: str = "theory",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
    as_numeric: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """
    Design power for binary metrics.

    Parameters
    ----------
    prob_a : float
       Probability of success for the control group.
    sizes : SampleSizeType
        List of single value of group sizes.
        For example: ``100``, ``[100, 200]``.
    effects : EffectType
        List or single value of relative effects.
        For example: ``1.05``, ``[1.05, 1.2]``.
    first_type_errors : StatErrorType, default: ``(0.05,)``
       I type error bounds
       P (detect difference for equal) < alpha.
    method: str, default: ``"theory"``
        Supports 2 methods: ``"theory"`` and ``"binary"``
        ``"theory"`` ~ by formula using statsmodels solve_power mechanism
        ``"binary"`` ~ using different types of intervals
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.
    as_numeric : bool, default: ``False``
        The result of calculations can be obtained as a percentage string
        either as a number, this parameter could used to toggle.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_table : pd.DataFrame
        Table with results of design.
    """
    if isinstance(effects, (int, float)):
        effects = [effects]
    if isinstance(sizes, int):
        sizes = [sizes]
    if isinstance(first_type_errors, float):
        first_type_errors = [first_type_errors]
    if method == "theory":
        return theory_pkg.get_power_table(
            mean=prob_a,
            std=None,
            sample_sizes=sizes,
            effects=effects,
            first_errors=first_type_errors,
            as_numeric=as_numeric,
            target_type="binary",
            groups_ratio=groups_ratio,
            alternative=alternative,
            stabilizing_method=stabilizing_method,
        )
    elif method == "binary":
        return bin_pkg.get_table_power_on_size_and_delta(
            p_a=prob_a,
            sample_sizes=sizes,
            first_errors=first_type_errors,
            delta_relative_values=effects,
            as_numeric=as_numeric,
            **kwargs,
        )
    else:
        raise ValueError(f"Choose valid method from {BINARY_DESIGN_METHODS}, got {method}")


def design_binary(
    to_design: str,
    prob_a: float,
    sizes: Optional[types.SampleSizeType] = None,
    effects: Optional[types.EffectType] = None,
    first_type_errors: types.StatErrorType = (0.05,),
    second_type_errors: types.StatErrorType = (0.2,),
    method: str = "theory",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
    **kwargs,
) -> pd.DataFrame:
    """
    Design of experiment parameters for binary metrics based
    on a known conversion value.

    Parameters
    ----------
    to_design : str
        Parameter to design.
    prob_a : float
        Probability of success for the control group.
    sizes : SampleSizeType, optional
        List or single value of group sizes.
        For example: ``100``, ``[100, 200]``.
    effects : EffectType, optional
        List of single value of relative effects.
        For example: 1.05, [1.05, 1.2].
    first_type_errors : StatErrorType, default: ``(0.05, )``
        I type error bounds
        P (detect difference for equal) < alpha.
    second_type_errors : StatErrorType, default: ``(0.2,)``
        II type error bounds
        P (suppose equality for different groups) < beta.
    method: str, default: ``"theory"``
        Supports 2 methods: ``"theory"`` and ``"binary"``
        ``"theory"`` ~ by formula using statsmodels solve_power mechanism
        ``"binary"`` ~ using different types of intervals
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_table : pd.DataFrame
        Table with results of design.
    """
    if to_design == SIZE:
        return design_binary_size(
            prob_a,
            effects,
            first_type_errors,
            second_type_errors,
            method,
            groups_ratio,
            alternative,
            stabilizing_method,
            **kwargs,
        )
    elif to_design == EFFECT:
        return design_binary_effect(
            prob_a,
            sizes,
            first_type_errors,
            second_type_errors,
            method,
            groups_ratio,
            alternative,
            stabilizing_method,
            **kwargs,
        )
    elif to_design == POWER:
        return design_binary_power(
            prob_a, sizes, effects, first_type_errors, method, groups_ratio, alternative, stabilizing_method, **kwargs
        )
    else:
        raise ValueError(f"Only {SIZE}, {EFFECT} and {POWER} parameters of the binary experiment could be designed.")



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

import numpy as np
import pandas as pd

from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark

# PySpark types


class PySparkStub:
    pass


if spark_installed():
    SparkSession = pyspark.sql.session.SparkSession
    SparkDataFrame = pyspark.sql.dataframe.DataFrame
    SparkColumn = pyspark.sql.column.Column
    SparkOrPandas = Union[SparkDataFrame, pd.DataFrame]
else:
    SparkSession = PySparkStub
    SparkDataFrame = PySparkStub
    SparkColumn = PySparkStub
    SparkOrPandas = PySparkStub


# Global types

# Type for passed data
PassedDataType = Union[pd.DataFrame, SparkDataFrame, str]

# Type for selected set of arguments names and values
_UsageArgumentsType = Dict

# Type of the unified sets of arguments prepared for the  further selection
_PrepareArgumentsType = Dict

# Type for columns name
ColumnNameType = str

# Type for columns
ColumnNamesType = Union[ColumnNameType, Iterable[ColumnNameType]]

# Type for metrics name
MetricNameType = ColumnNameType

# Type for metrics names
MetricNamesType = Union[MetricNameType, Iterable[MetricNameType]]

# Type for statistical errors
StatErrorType = Union[Iterable[float], float]

# Type for Statistical Criterion results
StatCriterionResult = Dict[str, Any]


# Tools types

# Type for bootstraped samples A/B
BootstrapedSamplesType = Dict[str, np.ndarray]

# Criterion result (statistic, pvalue)
CriterionResultType = Tuple[np.ndarray, np.ndarray]

# Type for one coinfedence interval
SingleIntervalType = Tuple[float, float]

# Coinfedence intervall type
ManyIntervalType = Tuple[np.ndarray, np.ndarray]

# Type for coinfedence interval - 1 or many
IntervalType = Union[SingleIntervalType, ManyIntervalType]

# Type for criterion
CompoundCriterionType = Union[Callable[[np.ndarray, np.ndarray], CriterionResultType], str]


# Designer types

# Type for size(s) of groups in A/B experiment
SampleSizeType = Union[Iterable[int], int]

# Type for effect(s) in A/B experiment
EffectType = Union[Iterable[float], float]

# Type for the result of designing the experiment via Designer
DesignerResult = Union[pd.DataFrame, Dict[MetricNameType, pd.DataFrame]]


# Splitter types

# Type for indices set of A or B group
IndicesType = Iterable[int]

# Type for the result of splitting data in A/B groups via Splitter
SplitterResult = Union[pd.DataFrame, SparkDataFrame]


# Tester types

# Type for data containg the information about group belonging of experimental objects
GroupsInfoType = pd.DataFrame

# Type for A and B groups data
TwoSamplesType = Tuple[pd.DataFrame, pd.DataFrame]

# Type for pair of labels belongs to A and B groups
GroupLabelsType = Tuple[Any, Any]

# Type for the values of metrics of experimental groups
GroupType = np.ndarray

# Type for experiment measurements packed in dictionary
ExperimentResults = Dict[str, PassedDataType]

# Type for experiment measurement action result for one metric
_SubResultType = Dict[str, Any]

# Type for the result of effect measurement in the A/B experiment via Tester
TesterResult = Union[pd.DataFrame, List[Dict[str, Any]]]


# AggregatePreprocessor types

# Type for set of methods/functions for data aggregation
MethodType = Union[str, Callable[[pd.Series], Any]]



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, List, Optional, Sequence

import numpy as np
import pandas as pd

import ambrosia.tools.empirical_tools as emp_pkg
from ambrosia import types

EPSILON: float = 0.001


def __helper_generate_bootstrap_samples(
    dataframe: pd.DataFrame, metrics: List[str], total_size: int, bootstrap_size: int, random_seed: Optional[int] = None
) -> types.BootstrapedSamplesType:
    """
    Make dictionary {metric : samples}.

    Samples - groups A/B (total_size, bootstrap_size)
    """
    rng = np.random.default_rng(random_seed)
    sampled_metrics: types.BootstrapedSamplesType = {}
    for metric in metrics:
        sampled_metric_vals = rng.choice(dataframe[metric].values, size=(total_size, bootstrap_size))
        sampled_metrics[metric] = sampled_metric_vals
    return sampled_metrics


def __helper_inject_effect(
    sampled_metrics: types.BootstrapedSamplesType,
    sample_size_a: int,
    effect: float,
    injection_method: str = "constant",
    variation_factor: float = 10,
    random_seed: Optional[int] = None,
) -> types.BootstrapedSamplesType:
    """
    Help to inject effect after sampling groups A/B.
    """
    modified_samples_values: types.BootstrapedSamplesType = {}
    for metric, sampled_metric in sampled_metrics.items():
        modified_samples_values[metric] = emp_pkg.inject_effect(
            sampled_metric,
            sample_size_a=sample_size_a,
            effect=effect,
            modeling_method=injection_method,
            variation_factor=variation_factor,
            random_seed=random_seed,
        )
    return modified_samples_values


def __helper_get_power_for_bootstraped(
    modified_samples: types.BootstrapedSamplesType,
    sample_size: int,
    bound_size: int,
    alpha: float,
    groups_ratio: float = 1.0,
    criterion: str = "ttest",
    random_seed: Optional[int] = None,
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> List[float]:
    """
    Calculate power for bootstraped samples.
    """
    result_power = []
    for _, values in modified_samples.items():
        sampled_metric_vals = np.vstack(
            [values[:sample_size], values[bound_size : bound_size + int(groups_ratio * sample_size)]]
        )
        power = emp_pkg.eval_error(
            sampled_metric_vals,
            sample_size_a=sample_size,
            alpha=alpha,
            mode=criterion,
            random_seed=random_seed,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        result_power.append(power)
    return result_power


def estimate_power(power_function: Callable, **kwargs_power) -> float:
    """
    Helps calc power with cases with cases with multioutput.
    """
    power_estimation = power_function(**kwargs_power)
    if isinstance(power_estimation, Sequence):
        power_estimation = power_estimation[0]
    return power_estimation


def helper_bin_search_upper_bound_size(
    power_function: Callable,
    power_level: float,
    groups_sizes_names: List[str],
    groups_ratio: float = 1.0,
    **kwargs_power,
) -> int:
    """
    Binary search for upper bound group size.
    """
    upper_bound_degree: int = 4
    power_estimation: float = 0
    while power_estimation < power_level:
        for gr_name in groups_sizes_names:
            kwargs_power[gr_name] = 2**upper_bound_degree
        kwargs_power[groups_sizes_names[-1]] = int(groups_ratio * kwargs_power[groups_sizes_names[-1]])
        power_estimation: float = estimate_power(power_function, **kwargs_power)
        upper_bound_degree += 1
    return upper_bound_degree


def helper_bin_searh_upper_bound_effect(power_function: Callable, power_level: float, **kwargs_power) -> int:
    """
    Binary search for upper bound effect.
    """
    upper_bound_degree: float = 1
    power_estimation: float = 0
    while power_estimation < power_level:
        kwargs_power["effect"] = 2**upper_bound_degree
        power_estimation = estimate_power(power_function, **kwargs_power)
        upper_bound_degree += 1
    return upper_bound_degree


def helper_binary_search_optimal_effect(
    power_function: Callable,
    power_level: float,
    upper_bound_effect: float,
    bootstraped_samples: np.ndarray,
    injection_method: str,
    epsilon: float = EPSILON,
    **kwargs_power,
) -> float:
    """
    Binary search for optimal effect for power and size.
    """
    left: float = 1
    right: float = upper_bound_effect
    while right - left > epsilon:
        middle: float = (left + right) / 2
        modified_samples = __helper_inject_effect(
            bootstraped_samples,
            sample_size_a=kwargs_power["sample_size"],
            effect=middle,
            injection_method=injection_method,
            random_seed=kwargs_power["random_seed"],
        )
        power_estimation: float = power_function(**kwargs_power, modified_samples=modified_samples)[0]
        if power_estimation >= power_level:
            right = middle
        else:
            left = middle
    return right


def helper_binary_search_effect_with_injection(
    power_function: Callable,
    power_level: float,
    upper_bound_effect: float,
    effect_injection_name: str,
    epsilon: float = EPSILON,
    **kwargs_power,
) -> float:
    """
    Binary search for effect using function incapsulating injection.
    """
    left: float = 1
    right: float = upper_bound_effect
    while right - left > epsilon:
        middle: float = (left + right) / 2
        kwargs_power[effect_injection_name] = middle
        power_estimation: float = estimate_power(power_function, **kwargs_power)
        if power_estimation >= power_level:
            right = middle
        else:
            left = middle
    return right


def helper_binary_search_optimal_size(
    power_function: Callable, power_level: float, upper_bound_size: int, groups_sizes_names: List[str], **kwargs_power
) -> int:
    """
    Binary search for optimal groups size.
    """
    left: int = 1
    right: int = upper_bound_size

    while right - left > 1:
        middle: int = (left + right) // 2
        for name in groups_sizes_names:
            kwargs_power[name] = middle
        power_estimation: float = estimate_power(power_function, **kwargs_power)
        if power_estimation >= power_level:
            right = middle
        else:
            left = middle
    return right



import contextlib
from typing import Any, Callable, Dict, Iterable, Optional, Union

import joblib
import numpy as np
from tqdm.auto import tqdm

from ambrosia import types
from ambrosia.tools.decorators import tqdm_parallel_decorator


def create_seed_sequence(length: int, entropy: Optional[Union[int, Iterable[int]]] = None) -> np.ndarray:
    """
    Create a seed sequence using ``numpy.random.SeedSequence`` class.
    Parameters
    ----------
    length : int
        Total length of a sequence.
    entropy : Union[int,Iterable[int]], optional
        The entropy for creating a ``SeedSequence``.
        Used to get a deterministic result.
    Returns
    -------
    seed_sequence : List
        The seed sequence of requested length.
    """
    rng = np.random.SeedSequence(entropy)
    seed_sequence: np.ndarray = rng.generate_state(length)
    return seed_sequence


from ambrosia import types


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """
    Context manager to patch joblib to report into tqdm progress bar given as argument.
    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):  # pylint: disable=W0235
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def handle_nested_multiprocessing(
    n_jobs: int, criterion: int, func: Callable, desc: str, total: int, **kwargs
) -> Dict[str, Any]:
    """
    Handle parameters for nested bootstrap parallelism of experiment design computation tasks.
    """
    progress_bar = tqdm(desc=desc, total=total)
    if criterion == "bootstrap":
        nested_n_jobs, n_jobs = n_jobs, 1
        func = tqdm_parallel_decorator(func)
        kwargs["progress_bar"] = progress_bar
    else:
        nested_n_jobs: int = 1
        progress_bar = tqdm_joblib(progress_bar)
    return {
        "n_jobs": n_jobs,
        "nested_n_jobs": nested_n_jobs,
        "parallel_func": func,
        "progress_bar": progress_bar,
        "kwargs": kwargs,
    }


def wrap_cols(cols: types.ColumnNamesType) -> types.ColumnNamesType:
    """
    Handle one or number of columns as list.
    """
    if isinstance(cols, types.ColumnNameType):
        cols = [cols]
    return cols



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import product
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from hyperopt import fmin, hp, tpe
from joblib import Parallel, delayed, parallel_backend

import ambrosia.tools._lib._tools_aide as aid_pkg
import ambrosia.tools.empirical_tools as emp_pkg
from ambrosia import types
from ambrosia.tools import back_tools

from . import EFFECT_COL_NAME, FIRST_TYPE_ERROR_COL_NAME, GROUP_SIZE_COL_NAME, STAT_ERRORS_COL_NAME

ROUND_DIGITS_TABLE: int = 3
ROUND_DIGITS_PERCENT: int = 1


def bootstrap_over_statistical_population(
    dataframe: pd.DataFrame,
    metrics: List,
    sample_size_a: int,
    sample_size_b: int,
    effect: float,
    alpha: float,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    n_jobs: bool = 1,
    verbose: bool = False,
    **kwargs,
) -> List[Optional[float]]:
    """
    Evaluate errors of the experiment setup for each metric and set of parameters.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe.
    metrics : List
        List of metrics - columns of ``dataframe``.
    sample_size_a : int
        Size of group A.
    sample_size_b : int
        Size of group B.
    effect : float
        Expected effect in group B for the modeling.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion to apply.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_list : List[Optional[float]]
        List of calculated correctness/power for bootstraped groups
        [corretness_0/power_0, correctness_1/power_1 ... corretness_k/power_k]
    """
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(metrics), random_seed)
    iterator = zip(metrics, seed_sequence)
    result_list = []
    for metric, seed in iterator:
        metric_vals = dataframe[metric].values.astype("float32")  ## to discuss
        sampled_metric_vals = np.random.default_rng(seed).choice(
            metric_vals, size=(sample_size_a + sample_size_b, bs_samples), replace=True
        )
        modified_metric_vals = emp_pkg.inject_effect(
            sampled_metric_vals, sample_size_a, effect, modeling_method=injection_method, random_seed=seed
        )
        power = emp_pkg.eval_error(
            modified_metric_vals,
            sample_size_a,
            alpha,
            mode=criterion,
            random_seed=seed,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        result_list.append(power)
    return result_list


def get_errors(
    dataframe: pd.DataFrame,
    metrics: List,
    sample_sizes_a: Iterable[int],
    sample_sizes_b: Iterable[int],
    effects: Iterable[float],
    alphas: Iterable[float],
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> Tuple[np.ndarray, List[np.ndarray]]:
    """
    Get I/II type errors estimation for the parameters of experiment.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metrics : List
        Columns of ``dataframe`` for calculating errors.
    sample_sizes_a : Iterable[int]
        List of A group sizes.
    sample_sizes_b : Iterable[int]
        List of B group sizes.
    effect : float
        Expected effect in group B for the modeling.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    beta : float
        Second type error, 1 - beta - power of criterion.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    parameters, empirical_errors : Tuple[Tuple, np.ndarray]
        Parameters - Cartesian product of all sample sizes, effects, alphas
        empirical_errors - Array of lists [corretness_0, power_0,
        correctness_1, power_1 ... corretness_k, power_k].
    """
    parameters = tuple(product(zip(sample_sizes_a, sample_sizes_b), effects, alphas))
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(parameters), random_seed)
    iterator = zip(parameters, seed_sequence)
    handled_params: Dict = back_tools.handle_nested_multiprocessing(
        n_jobs,
        criterion,
        bootstrap_over_statistical_population,
        desc="Empirical errors calculation",
        total=len(parameters),
        **kwargs,
    )
    with handled_params["progress_bar"]:
        with parallel_backend(n_jobs=handled_params["n_jobs"], backend="loky"):
            empirical_errors = Parallel(verbose=verbose)(
                delayed(handled_params["parallel_func"])(
                    dataframe=dataframe,
                    metrics=metrics,
                    sample_size_a=params[0][0],
                    sample_size_b=params[0][1],
                    effect=params[1],
                    alpha=params[2],
                    bs_samples=bs_samples,
                    criterion=criterion,
                    injection_method=injection_method,
                    random_seed=seed,
                    n_jobs=handled_params["nested_n_jobs"],
                    verbose=verbose,
                    **handled_params["kwargs"],
                )
                for params, seed in iterator
            )
    return parameters, empirical_errors


def get_empirical_table_power(
    dataframe: pd.DataFrame,
    metrics: List,
    sample_sizes_a: Iterable[int],
    sample_sizes_b: Optional[Iterable[int]],
    effects: Iterable[int],
    alphas: Iterable[int],
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    n_jobs: int = 1,
    as_numeric: bool = False,
    verbose: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """
    Create table with calculated power for given data and set of parameters.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metrics : str
        Columns of ``dataframe`` for calculating errors.
    sample_sizes_a : Iterable[int]
        List of A group sizes.
    sample_sizes_b : Optional[Iterable[int]]
        List of B group sizes, if None it will copy group A sizes.
    effects : Iterable[float]
        Expected effects in group B.
    alphas : Iterable[float]
        First type error bounds, 1 - alpha: correctness
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    parallel : bool, default: ``False``
        Whether to use parallel computing.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    as_numeric : bool, default False
        Whether to return a number or a string with percentages.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    report : pd.DataFrame
        Table with sizes for group A, group B, effects, erros and metrics names
        |A size | B size | effect | alpha | name_correctness| name_power| ...
    """
    if isinstance(sample_sizes_a, List):
        sample_sizes_a = np.array(sample_sizes_a)
    if sample_sizes_b is None:
        sample_sizes_b = np.copy(sample_sizes_a)
    elif isinstance(sample_sizes_b, List):
        sample_sizes_b = np.array(sample_sizes_b)
    parameters, emprical_errors = get_errors(
        dataframe=dataframe,
        metrics=metrics,
        sample_sizes_a=sample_sizes_a,
        sample_sizes_b=sample_sizes_b,
        effects=effects,
        alphas=alphas,
        bs_samples=bs_samples,
        criterion=criterion,
        injection_method=injection_method,
        random_seed=random_seed,
        n_jobs=n_jobs,
        verbose=verbose,
        **kwargs,
    )
    reports = {}
    for num, metric in enumerate(metrics):
        metric_errors: List = [errors_subset[num] for errors_subset in emprical_errors]
        report = pd.DataFrame(
            list(parameters), columns=[GROUP_SIZE_COL_NAME, EFFECT_COL_NAME, FIRST_TYPE_ERROR_COL_NAME]
        ).join(pd.DataFrame(metric_errors, columns=[STAT_ERRORS_COL_NAME]))
        report[EFFECT_COL_NAME] = (round((report[EFFECT_COL_NAME] - 1) * 100, ROUND_DIGITS_PERCENT)).astype(str) + "%"
        report = report.pivot_table(
            index=[FIRST_TYPE_ERROR_COL_NAME, EFFECT_COL_NAME],
            columns=GROUP_SIZE_COL_NAME,
            values=STAT_ERRORS_COL_NAME,
            sort=False,
        )
        report = report[list(zip(sample_sizes_a, sample_sizes_b))]
        if as_numeric:
            report = report.applymap(lambda x: round(x, ROUND_DIGITS_TABLE))
        else:
            report = report.applymap(lambda x: str(round(x * 100, ROUND_DIGITS_PERCENT)) + "%")
        if len(metrics) == 1:
            reports = report
        else:
            reports[metric] = report
    return reports


def optimize_group_size(
    dataframe: pd.DataFrame,
    metric: str,
    effect: float,
    alpha: float,
    power: float,
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    epsilon: float = 0.001,
    solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> int:
    """
    Find optimal group size for each metric on given parameters set.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metric : str
        Column of ``dataframe`` to calculate criterion.
    effect : float
        Expected effect in group B for the modeling.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    power : float
        Desired power of criterion.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals number for optimization.
    epsilon : float, default: ``0.001``
        Precision for power estimation solution.
    soltuion : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    optimal_group_size : int
        Optimal size for groups A/B to reach sufficient power and confidence
        levels.
    """
    solutions_names: List[str] = ["hyperopt", "binary"]
    power_calculation: Callable = bootstrap_over_statistical_population

    def objective(params: Dict) -> float:
        group_size_a = int(params["group_size"])
        group_size_b = int(groups_ratio * params["group_size"])
        power_emp = power_calculation(
            dataframe=dataframe,
            metrics=[metric],
            sample_size_a=group_size_a,
            sample_size_b=group_size_b,
            effect=effect,
            alpha=alpha,
            bs_samples=bs_samples,
            criterion=criterion,
            injection_method=injection_method,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        delta = abs(power_emp - (power + epsilon))
        return delta

    upper_bound_degree: int = aid_pkg.helper_bin_search_upper_bound_size(
        power_calculation,
        power,
        ["sample_size_a", "sample_size_b"],
        groups_ratio,
        dataframe=dataframe,
        metrics=[metric],
        effect=effect,
        alpha=alpha,
        bs_samples=bs_samples,
        criterion=criterion,
        injection_method=injection_method,
        random_seed=random_seed,
        n_jobs=n_jobs,
        verbose=verbose,
        **kwargs,
    )
    if solution == "hyperopt":
        lower_bound_degree: int = max(0, upper_bound_degree // 2)
        # log(2) for reduction to binary logarithm
        space = {
            "group_size": hp.qloguniform(
                "group_size", lower_bound_degree * np.log(2), upper_bound_degree * np.log(2), 1
            )
        }
        best = fmin(objective, space, algo=tpe.suggest, max_evals=evals, verbose=False)
        optimal_group_size = int(best["group_size"])
    elif solution == "binary":
        upper_bound: int = 2**upper_bound_degree
        total_size: int = int((1 + groups_ratio) * upper_bound)
        bootstraped_samples: types.BootstrapedSamplesType = aid_pkg.__helper_generate_bootstrap_samples(
            dataframe=dataframe,
            metrics=[metric],
            total_size=total_size,
            bootstrap_size=bs_samples,
            random_seed=random_seed,
        )

        modified_samples: types.BootstrapedSamplesType = aid_pkg.__helper_inject_effect(
            sampled_metrics=bootstraped_samples,
            sample_size_a=upper_bound,
            effect=effect,
            injection_method=injection_method,
            random_seed=random_seed,
        )
        optimal_group_size: int = aid_pkg.helper_binary_search_optimal_size(
            aid_pkg.__helper_get_power_for_bootstraped,
            power,
            upper_bound,
            ["sample_size"],
            modified_samples=modified_samples,
            bound_size=upper_bound,
            alpha=alpha,
            groups_ratio=groups_ratio,
            criterion=criterion,
            random_seed=random_seed,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
    else:
        raise ValueError(f'Choose soltuion from {", ".join(solutions_names)}')
    return optimal_group_size


def calculate_group_size(
    dataframe: pd.DataFrame,
    metrics: List,
    effect: float,
    alpha: float,
    beta: float,
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> np.ndarray:
    """
    Calculate optimal group_size for each metric on given set of parameters.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metrics : List
        Columns of ``dataframe`` for calculating sufficient group sizes.
    effect : float
        Expected effect in group B for the modeling.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    beta : float
        Second type error, 1 - beta - power of criterion.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals number for optimization.
    optim_soltuion : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    group_sizes : np.ndarray
        Optimal group size for each metric from metrics list
    """
    power: float = 1 - beta
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(metrics), random_seed)
    iterator = zip(metrics, seed_sequence)
    group_sizes: List = []
    for metric, seed in iterator:
        optimal_group_size = optimize_group_size(
            dataframe=dataframe,
            metric=metric,
            effect=effect,
            alpha=alpha,
            power=power,
            groups_ratio=groups_ratio,
            bs_samples=bs_samples,
            criterion=criterion,
            injection_method=injection_method,
            random_seed=seed,
            evals=evals,
            solution=optim_solution,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        group_sizes.append(optimal_group_size)
    return np.array(group_sizes)


def get_group_sizes(
    dataframe: pd.DataFrame,
    metrics: List,
    effects: Iterable[float],
    alphas: Iterable[float],
    betas: Iterable[float],
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> Tuple[Tuple, List[np.ndarray]]:
    """
    Get optimal group sizes for the possible parameters sets of the experiment.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metrics : List
        Columns of ``dataframe`` for calculating sufficient group sizes.
    effects : Iterable[float]
        Expected effects in group B for the modeling.
    alphas : Iterable[float]
        First type error bounds, 1 - alpha: correctness.
    betas : Iterable[float]
        Second type errors, 1 - beta = power.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals number for optimization.
    optim_soltuion : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    parameters, group_sizes_list : Tuple[Tuple, Tuple[np.ndarray]]
        Parameters - Cartesian product of all sample sizes, effects, alphas
        group_sizes_list - Array of lists of sizes for each metric from metrics list.
    """
    parameters: Tuple = tuple(product(effects, alphas, betas))
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(parameters), random_seed)
    iterator = zip(parameters, seed_sequence)
    handled_params: Dict = back_tools.handle_nested_multiprocessing(
        n_jobs, criterion, calculate_group_size, desc="Group sizes calculation", total=len(parameters), **kwargs
    )
    with handled_params["progress_bar"]:
        with parallel_backend(n_jobs=handled_params["n_jobs"], backend="loky"):
            group_sizes_list = Parallel(verbose=verbose)(
                delayed(handled_params["parallel_func"])(
                    dataframe=dataframe,
                    metrics=metrics,
                    effect=params[0],
                    alpha=params[1],
                    beta=params[2],
                    groups_ratio=groups_ratio,
                    criterion=criterion,
                    bs_samples=bs_samples,
                    injection_method=injection_method,
                    n_jobs=handled_params["nested_n_jobs"],
                    verbose=verbose,
                    evals=evals,
                    optim_solution=optim_solution,
                    random_seed=seed,
                    **handled_params["kwargs"],
                )
                for params, seed in iterator
            )
    return parameters, group_sizes_list


def get_empirical_table_sample_size(
    dataframe: pd.DataFrame,
    metrics: List,
    effects: Iterable[float],
    alphas: Iterable[float],
    betas: Iterable[float],
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """
    Create pandas table with optimal group sizes for given data and set of parameters.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Given datframe with data for experiment.
    metrics : List
        Columns of ``dataframe`` for calculating sufficient group sizes.
    effects : Iterable[float]
        Expected effects in group B.
    alphas : Iterable[float]
        First type error bounds, 1 - alpha: correctness.
    betas : Iterable[float]
        Second type errors, 1 - beta = power.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals amount for optimization.
    optim_solution : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    report : Union[pd.DataFrame, Dict[str, pd.DataFrame]
        Tables with sizes for group A, group B, effects, errors and metrics names
        Effects as indices
        (alpha(1 type error), beta(2 type error)) for columns
        table for each metric: dict[metric name] = table with sizes.
    """
    parameters, group_sizes_list = get_group_sizes(
        dataframe=dataframe,
        metrics=metrics,
        effects=effects,
        alphas=alphas,
        betas=betas,
        groups_ratio=groups_ratio,
        bs_samples=bs_samples,
        criterion=criterion,
        injection_method=injection_method,
        random_seed=random_seed,
        evals=evals,
        optim_solution=optim_solution,
        n_jobs=n_jobs,
        verbose=verbose,
        **kwargs,
    )
    reports = {}
    for num, metric in enumerate(metrics):
        report = pd.DataFrame(list(parameters), columns=[EFFECT_COL_NAME, "alpha", "beta"]).join(
            pd.DataFrame(list(np.array(group_sizes_list)[:, num]), columns=[GROUP_SIZE_COL_NAME])
        )
        report[STAT_ERRORS_COL_NAME] = tuple(zip(report["alpha"], report["beta"]))
        report = report.pivot_table(
            index=EFFECT_COL_NAME, columns=STAT_ERRORS_COL_NAME, values=GROUP_SIZE_COL_NAME, sort=False
        )
        report.index = (np.round((report.index - 1) * 100, ROUND_DIGITS_PERCENT)).astype(str) + "%"
        if len(metrics) == 1:
            reports = report
        else:
            reports[metric] = report
    return reports


def optimize_mde(
    dataframe: pd.DataFrame,
    metric: str,
    group_size: int,
    alpha: float = 0.05,
    power: float = 0.8,
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> float:
    """
    Find minimum detectable effect for each metric on given parameters set.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metric : str
        Column of ``dataframe`` to calculate criterion.
    group_size : int
        Size for group A/B.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    power : float
        Desired power of criterion.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals amount for optimization.
    solution : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    mde : float
        Minimal detectable effect for given power and group size.
    """
    solutions_names: List[str] = ["hyperopt", "binary"]
    power_calculation: Callable = bootstrap_over_statistical_population
    sample_size_a, sample_size_b = group_size, int(groups_ratio * group_size)

    def objective(params):
        effect = params["effect"]
        power_emp = power_calculation(
            dataframe=dataframe,
            metrics=[metric],
            sample_size_a=sample_size_a,
            sample_size_b=sample_size_b,
            effect=effect,
            alpha=alpha,
            bs_samples=bs_samples,
            criterion=criterion,
            injection_method=injection_method,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        if power_emp > power:
            val = effect
        else:
            # Some value more than right bound
            val = np.inf
        return val

    upper_bound_effect: float = 2 ** aid_pkg.helper_bin_searh_upper_bound_effect(
        power_calculation,
        power,
        dataframe=dataframe,
        metrics=[metric],
        alpha=alpha,
        sample_size_a=sample_size_a,
        sample_size_b=sample_size_b,
        bs_samples=bs_samples,
        criterion=criterion,
        injection_method=injection_method,
        random_seed=random_seed,
        n_jobs=n_jobs,
        verbose=verbose,
        **kwargs,
    )
    if solution == "hyperopt":
        space = {"effect": hp.uniform("effect", 1, upper_bound_effect)}
        best = fmin(objective, space, algo=tpe.suggest, max_evals=evals, verbose=False)
        optimal_effect: float = best["effect"]

    elif solution == "binary":
        bootstraped_samples: types.BootstrapedSamplesType = aid_pkg.__helper_generate_bootstrap_samples(
            dataframe=dataframe,
            metrics=[metric],
            total_size=sample_size_a + sample_size_b,
            bootstrap_size=bs_samples,
            random_seed=random_seed,
        )
        optimal_effect: int = aid_pkg.helper_binary_search_optimal_effect(
            aid_pkg.__helper_get_power_for_bootstraped,
            power,
            upper_bound_effect,
            bootstraped_samples,
            injection_method,
            bound_size=group_size,
            sample_size=group_size,
            alpha=alpha,
            groups_ratio=groups_ratio,
            criterion=criterion,
            random_seed=random_seed,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
    else:
        raise ValueError(f'Choose soltuion from {", ".join(solutions_names)}')
    return optimal_effect


def calculate_empirical_mde(
    dataframe: pd.DataFrame,
    metrics: List,
    group_size: int,
    alpha: float,
    beta: float,
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> np.ndarray:
    """
    Calculate empirical MDE for each metric on given set of parameters.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metric : str
        Column of ``dataframe`` to calculate criterion.
    group_size : int
        Size for group A/B.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    beta : float
        Second type error, 1 - beta - power of criterion.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals amount for optimization.
    optim_solution : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    mdes : np.ndarray
        List of minimal detectable effect for each metric from metrics list.
    """
    power: float = 1 - beta
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(metrics), random_seed)
    iterator = zip(metrics, seed_sequence)
    mdes: List = []
    for metric, seed in iterator:
        mde: float = optimize_mde(
            dataframe=dataframe,
            metric=metric,
            group_size=group_size,
            alpha=alpha,
            power=power,
            groups_ratio=groups_ratio,
            bs_samples=bs_samples,
            criterion=criterion,
            injection_method=injection_method,
            random_seed=seed,
            evals=evals,
            solution=optim_solution,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
        mdes.append(mde)
    return mdes


def get_empirical_mde(
    dataframe: pd.DataFrame,
    metrics: List,
    group_sizes: Iterable[int],
    alphas: Iterable[float],
    betas: Iterable[float],
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> Tuple[Tuple, List[np.ndarray]]:
    """
    Get empirical MDEs for the possible parameters sets of experiment.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Given datframe with data for experiment.
    metrics : str
        Column of ``dataframe`` to calculate minimal detectable effect.
    group_sizes : Iterable[int]
        Sizes for groups A/B.
    alphas : Iterable[float]
        First type error bounds, 1 - alpha: correctness
    betas : Iterable[float]
        Second type erros, 1 - beta = power.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals amount for optimization.
    optim_solution : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    parameters, empirical_errors : Tuple[Tuple, np.ndarray]
        Parameters - Cartesian product of all sample sizes, effects, alphas
        mde_list - Array of lists of mde lists for each metric from metrics list.
    """
    parameters: Tuple = tuple(product(group_sizes, alphas, betas))
    seed_sequence: np.ndarray = back_tools.create_seed_sequence(len(parameters), random_seed)
    iterator = zip(parameters, seed_sequence)
    handled_params: Dict = back_tools.handle_nested_multiprocessing(
        n_jobs, criterion, calculate_empirical_mde, desc="MDE calculation", total=len(parameters), **kwargs
    )
    with handled_params["progress_bar"]:
        with parallel_backend(n_jobs=handled_params["n_jobs"], backend="loky"):
            mde_list = Parallel(verbose=verbose)(
                delayed(handled_params["parallel_func"])(
                    dataframe=dataframe,
                    metrics=metrics,
                    group_size=params[0],
                    alpha=params[1],
                    beta=params[2],
                    groups_ratio=groups_ratio,
                    criterion=criterion,
                    bs_samples=bs_samples,
                    injection_method=injection_method,
                    random_seed=seed,
                    evals=evals,
                    optim_solution=optim_solution,
                    n_jobs=handled_params["nested_n_jobs"],
                    verbose=verbose,
                    **handled_params["kwargs"],
                )
                for params, seed in iterator
            )
    return parameters, mde_list


def get_empirical_mde_table(
    dataframe: pd.DataFrame,
    metrics: List,
    group_sizes: Iterable[int],
    alphas: Iterable[float],
    betas: Iterable[float],
    groups_ratio: float = 1.0,
    bs_samples: int = 1000,
    criterion: str = "ttest",
    injection_method: str = "constant",
    random_seed: Optional[int] = None,
    evals: int = 50,
    optim_solution: str = "binary",
    n_jobs: int = 1,
    as_numeric: bool = False,
    verbose: bool = False,
    **kwargs,
) -> Union[pd.DataFrame, Dict]:
    """
    Create pandas table with MDEs for given data and set of parameters.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Given datframe with data for experiment.
    metrics : List
        Columns of ``dataframe`` for calculating sufficient group sizes.
    group_sizes : Iterable[int]
        Sizes for groups A/B.
    alphas : Iterable[float]
        First type error bounds, 1 - alpha: correctness.
    betas : Iterable[float]
        Second type errors, 1 - beta = power.
    groups_ratio : float, default: ``1.0``
        Ratio between groups A and B.
    bs_samples : int, default: ``1000``
        Amount of bootstrap pairs A/B.
    criterion : str, default: ``"ttest"``
        Statistical criterion.
    injection_method : str, default: ``"constant"``
        Method to modify group B, for example
        constant: value = value * effect
        for more information inspect ``inject_effect`` function.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    evals : int, default: ``50``
        Evals amount for optimization.
    optim_solution : str, default: ``"binary"``
        Optimizer method.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    as_numeric : bool, default False
        Whether to return a number or a string with percentages.
    verbose : bool, default: ``False``
        Whether use logging or not.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    report: Union[pd.DataFrame, Dict[str, pd.DataFrame]
        Tables with sizes for group A, group B, effects, erros and metrics names
        Group sizes for indices
        (alpha(1 type error), beta(2 type error)) for columns
        table for each metric: dict[metric name] = table with mde(effects).
    """
    parameters, effects = get_empirical_mde(
        dataframe=dataframe,
        metrics=metrics,
        group_sizes=group_sizes,
        alphas=alphas,
        betas=betas,
        groups_ratio=groups_ratio,
        bs_samples=bs_samples,
        criterion=criterion,
        injection_method=injection_method,
        random_seed=random_seed,
        evals=evals,
        optim_solution=optim_solution,
        n_jobs=n_jobs,
        verbose=verbose,
        **kwargs,
    )
    reports = {}
    for num, metric in enumerate(metrics):
        report = pd.DataFrame(list(parameters), columns=[GROUP_SIZE_COL_NAME, "alpha", "beta"]).join(
            pd.DataFrame(list(np.array(effects)[:, num]), columns=[EFFECT_COL_NAME])
        )
        report[STAT_ERRORS_COL_NAME] = tuple(zip(report["alpha"], report["beta"]))
        report = report.pivot_table(
            index=GROUP_SIZE_COL_NAME, columns=STAT_ERRORS_COL_NAME, values=EFFECT_COL_NAME, sort=False
        )
        if as_numeric:
            report = report.applymap(lambda x: round(x, ROUND_DIGITS_TABLE))
        else:
            report = report.applymap(lambda x: str(round((x - 1) * 100, ROUND_DIGITS_PERCENT)) + "%")
        report = report[[(alpha, beta) for alpha in alphas for beta in betas]]
        if len(metrics) == 1:
            reports = report
        else:
            reports[metric] = report
    return reports



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd

import ambrosia.tools.ab_abstract_component as ab_abstract
from ambrosia.tools import log


class Stratification(ab_abstract.StratificationUtil):
    """
    Stratification implementation
    https://en.wikipedia.org/wiki/Stratified_sampling

    Attributes
    ----------
    dataframe: pd.DataFrame
        Given data frame for stratification
    columns: List[Any]
        Columns for stratification
    threshold: Optional[int]
        If amount for current value of stratification is <= than threshold
        It wont be used for stratification. If None such regularization.
    verbose: bool
        Whenether to print information of not used strats
    strats: Dict[Any, pd.DataFrame]
        Dictionary with startification tables, values for current strat -> table

    Methods
    -------
    fit(dataframe: pd.DataFrame, columns: List[Any]) -> None
        Store data frame and stratification data
    strat_sizes() -> Dict[Any, int]
        Calculate stratification sizes
    get_test_inds(self, test_id: Iterable, id_column: Any=None) -> Dict[Any, Tuple[List, int]]
        Find test group ids for each stratification group and size of rest a group
    is_trained() -> bool
        Return true if fit method was called before
    is_not_trained() -> bool
        Return true if fit method was not called before
    size() -> int
        Return size considering droped rows using threshold in stratification
        If threshold = None => size() = len(dataframe)
    groups()
        Returns items for stratification
    """

    def __init__(self, threshold: Optional[int] = None, verbose: bool = False):
        """
        Parameters
        ----------
        threshold : Optional[int]
            Threshold for stratification group sizes
        verbose : bool, default: ``False``
            Whenether to print information of not used strats
        """
        super().__init__()
        self.dataframe = None
        self.columns = None
        self.threshold = threshold
        self.verbose = verbose

    def fit(self, dataframe: pd.DataFrame, columns: Optional[List[Any]] = None) -> None:
        """
        Store stratification data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Pandas Data Frame for stratification
        columns : Optional[List[Any]]
            Columns of dataframe for straitification
        """
        self.dataframe = dataframe
        self.columns = columns
        if columns is None:
            self.strats = {ab_abstract.EmptyStratValue.NO_STRATIFICATION: dataframe}
            return
        self.strats: Dict[Any, pd.DataFrame] = {}
        for values, table in dataframe.groupby(columns):
            if self.threshold is not None and self.dataframe.shape[0] <= self.threshold:
                if self.verbose:
                    log.info_log("Stratification group with values:")
                    for val, name in zip(values, columns):
                        log.info_log(f"Column {name}: {val}")
                continue
            self.strats[values] = table

    def strat_sizes(self) -> Dict[Any, int]:
        """
        Returns sizes for each strat
        """
        self._check_fit()
        sizes: Dict[Any, int] = {}
        for value, table in self.strats.items():
            sizes[value] = table.shape[0]
        return sizes

    @staticmethod
    def __corresponding_strat(test_id: Iterable, strat_id: Iterable) -> List:
        """
        Filter test id for given stratification group.
        """
        return list(filter(lambda x: x in strat_id, test_id))

    def get_test_inds(self, test_id: Iterable, id_column: Any = None) -> Dict[Tuple, Tuple[List, int]]:
        """
        Returns test ids for each strat and amount of remaining ids in this strat.
        Basically use id of data frame, if id_column is set use it

        Parameters
        ----------
        test_id : Iterable
            Ids for test group
        id_column : Any, default: None

        Returns
        -------
        test_ids : Dict[Tuple, Tuple[List, int]]
            Dictionary for each value from stratification with test ids and
            not used ids from this strat
        """
        if self.empty_strat():
            other_amount: int = self.size() - len(test_id)
            return {ab_abstract.EmptyStratValue.NO_STRATIFICATION: (list(test_id), other_amount)}

        if id_column is not None:
            error_column_name: str = f"""Column - {id_column},
             is not in list of columns - {", ".join(self.dataframe.columns)}"""
            if id_column not in self.dataframe.columns:
                raise ValueError(error_column_name)
        test_ids: Dict[Tuple, Tuple[List, int]] = {}
        for value, table in self.strats.items():
            if id_column is not None:
                ids: List = Stratification.__corresponding_strat(test_id, table[id_column].values)
            else:
                ids: List = Stratification.__corresponding_strat(test_id, table.index.values)
            test_ids[value] = (ids, table.shape[0] - len(ids))
        return test_ids



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Experiment results evaluation and testing methods.

Module contains `Tester` core class and `test` methid which are
designed to evaluate statistical significance of the experiment results
and a magnitude of effect via large number of methods and criteria.

It is recommended to use for parameters such as test method and statistical
criterion the values that were chosen during the experiment design stage.

Currently, experimental results can only be processed and evaluated as
pandas DataFrames or .csv tables. Support for Spark dataframes is under
development and will be available soon.

"""
import itertools
from copy import deepcopy
from typing import Callable, Dict, List, Optional, Union
from warnings import warn

import numpy as np
import pandas as pd

import ambrosia.tools.empirical_tools as empirical_pkg
import ambrosia.tools.pvalue_tools as pvalue_pkg
import ambrosia.tools.stat_criteria as criteria_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import ABStatCriterion, ABToolAbstract, DataframeHandler, StatCriterion

from .binary_result_evaluation import binary_absolute_result, binary_relative_result
from .handlers import TheoreticalTesterHandler, filter_spark_and_make_groups

BOOTSTRAP_SIZE: int = 10000
AVAILABLE: List[str] = ["pandas", "spark"]
AVAILABLE_AB_CRITERIA: Dict[str, ABStatCriterion] = {
    "ttest": criteria_pkg.TtestIndCriterion,
    "ttest_rel": criteria_pkg.TtestRelCriterion,
    "mw": criteria_pkg.MannWhitneyCriterion,
    "wilcoxon": criteria_pkg.WilcoxonCriterion,
}
AVAILABLE_MULTITEST_CORRECTIONS: List[str] = ["bonferroni"]


class Tester(ABToolAbstract):
    """
    Unit for evaluating the results of experiments.

    The experiment evaluation result contains:
        - Pvalue for the selected criterion
        - Point effect estimation
        - Corresponding confidence interval for the effect
        - Boolean result - presence / absence of the effect

    Parameters
    ----------
    dataframe : PassedDataType, optional
        Dataframe used with experiment results metrics.
    df_mapping : GroupsInfoType, optional
        Dataframe which contains group labels of objects.
    experiment_results : ExperimentResults, optional
        Dict with separate experiment results for each group.
        Dict keys are used as groups labels, values must be either
        pandas or Spark dataframes.
    column_groups : ColumnNameType, optional
        Column which contains groups label of objects.
    group_labels : GroupLabelsType, optional
        Labels for experimental groups. If ``column_groups`` contains
        at least two values, they will choose for labels.
    id_column : ColumnNameType, optional
        Name of column with objects ids in ``df_mapping`` dataframe.
    first_type_errors : StatErrorType, default: ``0.05``
        I type errors values. Fix P (detect difference for equal) to be less
        than threshold. Used to construct confidence intervals.
    metrics : MetricNameType, optional
        Metrics (columns of dataframe) which is used to calculate
        experiment result.
    metric_funcs : Dict[str, Callable], optional
        Dictionary mapping metric names to callable functions.
        Each function receives a ``pd.DataFrame`` (group data) and must
        return an array-like of numeric values. When provided, the
        function is used instead of column lookup for the corresponding
        metric name. Only supported for pandas DataFrames.

    Attributes
    ----------
    dataframe : PassedDataType
        Dataframe used with experiment results metrics.
    df_mapping : GroupsInfoType
        Dataframe which contains group labels of objects.
    experiment_results : ExperimentResults, optional
        Dict with separate experiment results for each group.
    column_groups : ColumnNameType
        Column which contains groups label of objects.
    group_labels : GroupLabelsType
        Labels for experimental groups.
    id_column : ColumnNameType
        Name of column with objects ids in ``df_mapping`` dataframe.
    first_type_errors : StatErrorType, default: ``0.05``
        I type errors values.
    metrics : MetricNameType
        Columns of dataframe with experiment results.

    Examples
    --------
    We've experimented with adding onboarding to our mobile app and
    would like to know about its results in terms of A/B testing.
    Suppose we have a loaded pandas dataframe with a column responsible
    for the groups in the testing and columns with metric values,
    such as retention. Then you can use the tester class the following way:

    >>> tester = Tester(
    >>>     dataframe=df,
    >>>     column_groups='groups',
    >>>     metrics='retention'
    >>> )
    >>> tester.run()
    >>> # Output
    >>> [{
    >>>     'first_type_error' : 0.05,
    >>>     'pvalue' : 0.03,
    >>>     'effect' : 1.05,
    >>>     'confidence_interval' : (1.01, 1.10),
    >>>     'metric name': 'retention',
    >>>     'group A label': 'A',
    >>>     'group B label': 'B'
    >>> }]

    Notes
    -----
    Basic mathematic methods for evaluating experiments:

        - Theory:
            - Absolute: Using ttest, mann-whitney, others and custom criteria
            - Relative: Using delta method

        - Empiric:
            - Absolute / Relative: Building empirical distribution for T(A, B)

        - Binary:
            - Absolute: Using special binary intervals and
              finding pvalue = inf_a {x : 0 not in interval(x)}
            - Relative: Not implemented yet :(

    Constructors:

    >>> # Empty constructor
    >>> tester = Tester()
    >>> # You can pass Iterable or single object for some parameters
    >>> tester = Tester(
    >>>     dataframe=df,
    >>>     columns_groups='groups',
    >>>     metrics=['ltv', 'retention']
    >>> )
    >>> tester = Tester(metrics='retention', first_type_errors=[0.01, 0.05])
    >>> # You can set a separate table containing information about
    >>> # the partitioning in the experiment
    >>> tester = tester = Tester(
    >>>     dataframe=df, # main dataframe with metrics
    >>>     df_mapping=groups, # table with information about groups
    >>>     metrics='metric', # Metric to be tested
    >>>     column_groups='group', # Column in df_mapping with labels
    >>>     id_column='id' # Column with ids in df and df_mapping (for join)
    >>> )

    Setters:

    >>> tester.set_metrics(['ltv', 'retention'])
    >>> tester.set_dataframe(dataframe=dataframe, column_groups='groups')
    >>> # You can set separate data of each group packed in special dict form
    >>> tester.set_experiment_results(experiment_results=experiment_results)

    Run:

    >>> # You can choose effect_type to estimate: relative / absolute
    >>> tester.run('absolute')
    >>> # Also you can choose method
    >>> tester.run('absolute', method='empriric') # emipiric for bootstrap
    >>> # One can pass arguments in run() method and they will have
    >>> # higher priority
    >>> tester.run(metrics='ltv', data_a_group=df_a)

    Use a function instead of a class:

    >>> test('absolute', dataframe=df, column_groups='groups', metrics='ltv')
    """

    # This is for avoiding warnings from pytest
    __test__ = False

    def set_experiment_results(self, experiment_results: types.ExperimentResults) -> None:
        self.__experiment_results = experiment_results

    def set_errors(self, first_type_errors: types.StatErrorType) -> None:
        if isinstance(first_type_errors, float):
            self.__alpha = np.array([first_type_errors])
        else:
            self.__alpha = np.array(first_type_errors)

    def set_metrics(self, metrics: types.MetricNamesType) -> None:
        if isinstance(metrics, types.MetricNameType):
            self.__metrics = [metrics]
        else:
            self.__metrics = metrics

    def set_dataframe(
        self,
        dataframe: types.PassedDataType,
        column_groups: types.MetricNameType,
        group_labels: types.GroupLabelsType = None,
        df_mapping: types.GroupsInfoType = None,
        id_column: types.MetricNameType = None,
    ) -> None:
        __filtering_kwargs = {
            "dataframe": dataframe,
            "df_mapping": df_mapping,
            "column_groups": column_groups,
            "group_labels": group_labels,
            "id_column": id_column,
        }
        self.__experiment_results = DataframeHandler()._handle_cases(
            Tester.__filter_data,
            filter_spark_and_make_groups,
            **__filtering_kwargs,
        )

    def __init__(
        self,
        dataframe: Optional[types.PassedDataType] = None,
        df_mapping: Optional[types.GroupsInfoType] = None,
        experiment_results: Optional[types.ExperimentResults] = None,
        column_groups: Optional[types.ColumnNameType] = None,
        group_labels: Optional[types.GroupLabelsType] = None,
        id_column: Optional[types.ColumnNameType] = None,
        first_type_errors: types.StatErrorType = 0.05,
        metrics: Optional[types.MetricNamesType] = None,
        metric_funcs: Optional[Dict[str, Callable]] = None,
    ):
        """
        Tester class constructor to initialize the object.
        """
        if dataframe is not None:
            self.set_dataframe(
                dataframe,
                column_groups,
                group_labels,
                df_mapping,
                id_column,
            )
        else:
            self.set_experiment_results(experiment_results=experiment_results)
        self.set_errors(first_type_errors)
        self.set_metrics(metrics)
        self.__metric_funcs = metric_funcs or {}

    @staticmethod
    def __filter_data(
        dataframe: types.PassedDataType,
        df_mapping: types.GroupsInfoType,
        column_groups: types.ColumnNameType,
        group_labels: types.GroupLabelsType,
        id_column: types.ColumnNameType,
    ) -> types.TwoSamplesType:
        """
        Function to handle setting of pandas data.
        """
        if dataframe is None:
            return None

        if df_mapping is not None:
            if id_column not in dataframe:
                raise ValueError(f"Column {id_column}, is not in list of df columns")
            if id_column not in df_mapping:
                raise ValueError(f"Column {id_column}, is not in list of df_mapping columns")
            dataframe = dataframe.merge(df_mapping, how="left", on=id_column).dropna()
        if column_groups not in dataframe:
            raise ValueError(f"Column {column_groups}, is not in list of df columns")

        if group_labels is not None:
            if len(group_labels) < 2:
                raise ValueError(f"Group labels must be at least 2, given {group_labels}")
        else:
            group_labels = dataframe[column_groups].unique()
        experiment_results: types.ExperimentResults = {
            group_label: dataframe[dataframe[column_groups] == group_label] for group_label in group_labels
        }
        return experiment_results

    @staticmethod
    def __bootstrap_result(
        group_a: types.GroupType,
        group_b: types.GroupType,
        alpha: np.ndarray,
        bootstrap_size: int = BOOTSTRAP_SIZE,
        effect_type: str = "absolute",
        **kwargs,
    ) -> types._SubResultType:
        """
        Function to handle the empirical approach to testing.
        """
        if effect_type == "absolute":
            metric = "mean"
            point_effect = np.mean(group_b) - np.mean(group_a)
        elif effect_type == "relative":
            metric = "fraction"
            point_effect = np.mean(group_b) / np.mean(group_a) - 1
        else:
            raise ValueError("Set effect_type as 'absolute' or 'relative'")
        paired: bool = kwargs.pop("paired") if "paired" in kwargs else False
        bootstrap_handler = empirical_pkg.BootstrapStats(bootstrap_size=bootstrap_size, metric=metric, paired=paired)
        bootstrap_handler.fit(group_a, group_b, **kwargs)
        left_bounds, right_bounds = bootstrap_handler.confidence_interval(confidence_level=1 - alpha, **kwargs)
        pvalue = bootstrap_handler.pvalue_criterion(**kwargs)
        confidence_interval = list(zip(left_bounds, right_bounds))
        return {
            "first_type_error": alpha,
            "pvalue": pvalue,
            "effect": point_effect,
            "confidence_interval": confidence_interval,
        }

    @staticmethod
    def __binary_result(
        group_a: types.GroupType, group_b: types.GroupType, alpha: np.ndarray, effect_type: str = "absolute", **kwargs
    ) -> types._SubResultType:
        """
        Function to handle binary intervals for testing.
        """
        warning_message_values: str = "Values for metric is not binary, choose other method, for example ttest!"
        if not set(np.unique(group_a)).issubset({0, 1}) or not set(np.unique(group_b)).issubset({0, 1}):
            warn(warning_message_values)
        if effect_type == "absolute":
            return binary_absolute_result(group_a, group_b, alpha, **kwargs)
        elif effect_type == "relative":
            return binary_relative_result(group_a, group_b, alpha, **kwargs)
        else:
            raise ValueError(f"``effect_type`` variable could be only  'absolute' or 'relative, got {effect_type}.")

    @staticmethod
    def __theory_handler(
        group_a: types.GroupType,
        group_b: types.GroupType,
        alpha: np.ndarray,
        effect_type: str = "absolute",
        criterion: Optional[ABStatCriterion] = None,
        **kwargs,
    ) -> types._SubResultType:
        """
        Function to handle the theoretical approach to testing.
        """
        criterion: Union[str, StatCriterion] = criterion if criterion is not None else "ttest"
        if isinstance(criterion, str) and (criterion in AVAILABLE_AB_CRITERIA):
            criterion = AVAILABLE_AB_CRITERIA[criterion]
        elif not (hasattr(criterion, "get_results") and callable(criterion.get_results)):
            raise ValueError(
                f"Choose correct criterion name from {list(AVAILABLE_AB_CRITERIA)} or pass correct custom class"
            )
        return criterion().get_results(group_a=group_a, group_b=group_b, alpha=alpha, effect_type=effect_type, **kwargs)

    @staticmethod
    def __pre_run(method: str, args: types._UsageArgumentsType, **kwargs) -> types.TesterResult:
        """
        Function to handle run method on pandas dataframes.
        """
        # TODO: add methods to enum
        accepted_methods: List[str] = ["theory", "empiric", "binary"]
        if method not in accepted_methods:
            raise ValueError(f'Choose method from {", ".join(accepted_methods)}')
        result: types.TesterResult = {}
        metric_funcs: Dict = args.get("metric_funcs", {})
        for metric in args["metrics"]:
            metric_func = metric_funcs.get(metric)
            if metric_func is not None:
                a_values: np.ndarray = np.asarray(metric_func(args["data_a_group"]))
                b_values: np.ndarray = np.asarray(metric_func(args["data_b_group"]))
            else:
                a_values = args["data_a_group"][metric].values
                b_values = args["data_b_group"][metric].values
            if method == "theory":
                # TODO: Make it SolverClass ~ method
                # solver = SolverClass(...)
                # sub_result = solver.solve()
                solver = TheoreticalTesterHandler(
                    args["data_a_group"],
                    args["data_b_group"],
                    column=metric,
                    alpha=np.array(args["alpha"]),
                    effect_type=args["effect_type"],
                    criterion=args["criterion"],
                    metric_func=metric_func,
                    **kwargs,
                )
                sub_result = solver.solve()
            elif method == "empiric":
                sub_result = Tester.__bootstrap_result(
                    a_values, b_values, np.array(args["alpha"]), effect_type=args["effect_type"], **kwargs
                )
            elif method == "binary":
                sub_result = Tester.__binary_result(
                    a_values, b_values, np.array(args["alpha"]), effect_type=args["effect_type"], **kwargs
                )
            result[metric] = sub_result
        return result

    @staticmethod
    def __apply_first_stage_multitest_correction(
        alphas: types.StatErrorType, hypothesis_num: int, method: str = "bonferroni"
    ) -> types.StatErrorType:
        """
        Apply first stage of multitest correction for first type errors.
        """
        alphas = alphas.copy()
        if method == "bonferroni":
            alphas /= hypothesis_num
        return alphas

    @staticmethod
    def __apply_second_stage_multitest_correction(
        result: types.TesterResult, hypothesis_num: int, method: str = "bonferroni"
    ):
        """
        Apply second stage of multitest correction.
        """
        if method == "bonferroni":
            result["pvalue"] = (result["pvalue"].values * hypothesis_num).clip(max=1)
            result["first_type_error"] *= hypothesis_num
        return result

    @staticmethod
    def as_table(dict_result: types.TesterResult) -> pd.DataFrame:
        """
        Transform dict type output result to pandas DataFrame format.

        Parameters
        ----------
        dict_result : TesterResult
           Tester result as a dictionary.

        Returns
        -------
        result_table : pd.DataFrame
           Table with results.
        """
        answer: List[pd.DataFrame] = []
        for single_test in dict_result:
            metrics_names = list(dict_result[single_test].keys())
            metrics_names.remove("group_a_label")
            metrics_names.remove("group_b_label")
            for metric_name in metrics_names:
                tmp = deepcopy(dict_result[single_test][metric_name])
                tmp["metric name"] = metric_name
                tmp["group A label"] = dict_result[single_test]["group_a_label"]
                tmp["group B label"] = dict_result[single_test]["group_b_label"]
                if tmp["confidence_interval"][0][0] is not None:
                    tmp["confidence_interval"] = [
                        (round(left, Tester._PRECISION_DIGITS), round(right, Tester._PRECISION_DIGITS))
                        for left, right in tmp["confidence_interval"]
                    ]
                answer.append(pd.DataFrame(tmp))
        result_table = pd.concat(answer).reset_index(drop=True)
        return result_table

    def run(
        self,
        effect_type: str = "absolute",
        method: str = "theory",
        dataframe: Optional[types.PassedDataType] = None,
        df_mapping: Optional[types.GroupsInfoType] = None,
        experiment_results: Optional[types.ExperimentResults] = None,
        id_column: Optional[str] = None,
        column_groups: Optional[str] = None,
        group_labels: Optional[types.GroupLabelsType] = None,
        metrics: Optional[types.MetricNamesType] = None,
        first_type_errors: Optional[types.StatErrorType] = None,
        criterion: Optional[ABStatCriterion] = None,
        correction_method: Union[str, None] = "bonferroni",
        as_table: bool = True,
        metric_funcs: Optional[Dict[str, Callable]] = None,
        **kwargs,
    ) -> types.TesterResult:
        """
        The main method for testing and evaluating experimental results.

        Parameters
        ----------
        effect_type : str, default: ``"absolute"``
           Effect type to calculate.
           Could be ``"absolute"`` or ``"relative"``.
        method : str, default: ``"theory"``
           Type of testing approach.
           Can take the values ``"theory"``, ``"empiric"`` or ``"binary"``.
        dataframe : PassedDataType, optional
           Data used to calculate the results of an experiment.
        df_mapping : GroupsInfoType, optional
           Dataframe which contains group labels of objects.
        experiment_results : ExperimentResults
            Dict with separate experiment results for each group.
            Dict keys are used as groups labels, values must be either
            pandas or Spark dataframes.
        column_groups : ColumnNameType
            Column which contains groups label of objects.
        group_labels : GroupLabelsType
            Labels for experimental groups.
        id_column : ColumnNameType
            Name of column with objects ids in ``df_mapping`` dataframe.
        first_type_errors : StatErrorType, default: ``0.05``
            I type errors values.
        metrics : MetricNameType
            Columns of dataframe with experiment results.
        criterion : ABStatCriterion, optional
            Statistical criterion for hypotheses testing.
            If ``method`` is ``"theory"`` and no criterion provided,
            ttest for independent samples will be used.
        correction_method : Union[str, None], default: ``bonferroni``
            Method for pvalues and confidence intervals multitest correction.
            Total number of hypothesis is equal to the number of
            variants combinations * number of metrics passed.
        as_table : bool, default: ``True``
            Return the test results as a pandas dataframe.
            If ``False``, a list of dicts with results will be returned.
        metric_funcs : Dict[str, Callable], optional
            Dictionary mapping metric names to callable functions.
            Each function receives a group ``pd.DataFrame`` and returns
            array-like values. Overrides functions set in constructor
            for matching metric names. Only pandas DataFrames supported.
        **kwargs : Dict
            Other keyword arguments.

        Returns
        -------
        result : types.TesterResult
            Experiment results as pandas table or list of dicts for each metric
            and first type error.
        """
        if isinstance(metrics, types.MetricNameType):
            metrics = [metrics]
        if first_type_errors is not None:
            if isinstance(first_type_errors, float):
                first_type_errors = np.array([first_type_errors])
            else:
                first_type_errors = np.array(first_type_errors)
        if "alternative" in kwargs:
            pvalue_pkg.check_alternative(kwargs["alternative"])
        else:
            kwargs["alternative"] = "two-sided"

        __filtering_kwargs = {
            "dataframe": dataframe,
            "df_mapping": df_mapping,
            "column_groups": column_groups,
            "group_labels": group_labels,
            "id_column": id_column,
        }
        if dataframe is not None:
            experiment_results = DataframeHandler()._handle_cases(
                Tester.__filter_data, filter_spark_and_make_groups, **__filtering_kwargs
            )

        arguments_choice: types._PrepareArgumentsType = {
            "experiment_results": (self.__experiment_results, experiment_results),
            "metrics": (self.__metrics, metrics),
            "alpha": (self.__alpha, first_type_errors),
        }
        chosen_args: types._UsageArgumentsType = Tester._prepare_arguments(arguments_choice)
        chosen_args["effect_type"] = effect_type
        chosen_args["criterion"] = criterion
        effective_metric_funcs = {**self.__metric_funcs, **(metric_funcs or {})}
        chosen_args["metric_funcs"] = effective_metric_funcs

        hypothesis_num: int = len(list(itertools.combinations(chosen_args["experiment_results"], 2))) * len(
            chosen_args["metrics"]
        )
        if correction_method is not None and hypothesis_num > 1:
            if correction_method in AVAILABLE_MULTITEST_CORRECTIONS:
                chosen_args["alpha"] = Tester.__apply_first_stage_multitest_correction(
                    chosen_args["alpha"], hypothesis_num, correction_method
                )
            else:
                raise ValueError(f"Choose correction method from {AVAILABLE_MULTITEST_CORRECTIONS}")

        result: types.TesterResult = {}
        # Variating over all pairs of groups - comb(n, 2)
        for group_a_label, group_b_label in itertools.combinations(chosen_args["experiment_results"], 2):
            test_name = f"group_{group_a_label}_vs_group_{group_b_label}"
            chosen_args["data_a_group"] = chosen_args["experiment_results"][group_a_label]
            chosen_args["data_b_group"] = chosen_args["experiment_results"][group_b_label]
            pre_run_args = (method, chosen_args)
            subresult: types.TesterResult = Tester.__pre_run(*pre_run_args, **kwargs)
            subresult["group_a_label"] = group_a_label
            subresult["group_b_label"] = group_b_label
            result[test_name] = subresult

        result = Tester.as_table(result)
        if correction_method is not None and hypothesis_num > 1:
            result = Tester.__apply_second_stage_multitest_correction(result, hypothesis_num, correction_method)
        if not as_table:
            result = result.to_dict(orient="records")
        return result


def test(
    effect_type: str = "absolute",
    method: str = "theory",
    dataframe: Optional[types.PassedDataType] = None,
    df_mapping: Optional[types.GroupsInfoType] = None,
    experiment_results: Optional[types.ExperimentResults] = None,
    id_column: Optional[str] = None,
    column_groups: Optional[str] = None,
    group_labels: Optional[types.GroupLabelsType] = None,
    metrics: Optional[types.MetricNamesType] = None,
    first_type_errors: Optional[types.StatErrorType] = None,
    criterion: Optional[ABStatCriterion] = None,
    correction_method: Union[str, None] = "bonferroni",
    as_table: bool = True,
    metric_funcs: Optional[Dict[str, Callable]] = None,
    **kwargs,
) -> types.TesterResult:
    """
    Function wrapper around the ``Tester`` class.

    Apply on the experimental data to get the results of an experiment.

    Creates an instance of the ``Tester`` class internally and execute
    run method with corresponding arguments.

    Parameters
    ----------
    effect_type : str, default: ``"absolute"``
        Effect type to calculate.
        Could be ``"absolute"`` or ``"relative"``.
    method : str, default: ``"theory"``
        Type of testing approach.
        Can take the values ``"theory"``, ``"empiric"`` or ``"binary"``.
    dataframe : PassedDataType, optional
        Data used to calculate the results of an experiment.
    df_mapping : GroupsInfoType, optional
        Dataframe which contains group labels of objects.
    experiment_results : ExperimentResults
        Dict with separate experiment results for each group.
        Dict keys are used as groups labels, values must be either
        pandas or Spark dataframes.
    column_groups : ColumnNameType
        Column which contains groups label of objects.
    group_labels : GroupLabelsType
        Labels for experimental groups.
    id_column : ColumnNameType
        Name of column with objects ids in ``df_mapping`` dataframe.
    first_type_errors : StatErrorType, default: ``0.05``
        I type errors values.
    metrics : MetricNameType
        Columns of dataframe with experiment results.
    criterion : ABStatCriterion, optional
        Statistical criterion for hypotheses testing.
        If ``method`` is ``"theory"`` and no criterion provided,
        ttest for independent samples will be used.
    correction_method : Union[str, None], default: ``bonferroni``
        Method for pvalues and confidence intervals multitest correction.
        Total number of hypothesis is equal to the number of
        variants combinations * number of metrics passed.
    as_table : bool, default: ``True``
        Return the test results as a pandas dataframe.
        If ``False``, a list of dicts with results will be returned.
    metric_funcs : Dict[str, Callable], optional
        Dictionary mapping metric names to callable functions.
        Each function receives a group ``pd.DataFrame`` and returns
        array-like values. Only pandas DataFrames supported.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result : types.TesterResult
        Experiment results as pandas table or list of dicts for each metric
        and first type error.
    """
    return Tester(
        dataframe=dataframe,
        df_mapping=df_mapping,
        id_column=id_column,
        column_groups=column_groups,
        group_labels=group_labels,
        metrics=metrics,
        first_type_errors=first_type_errors,
    ).run(
        effect_type=effect_type,
        method=method,
        experiment_results=experiment_results,
        criterion=criterion,
        correction_method=correction_method,
        as_table=as_table,
        metric_funcs=metric_funcs,
        **kwargs,
    )



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, List

import ambrosia.tools._lib._tools_aide as pkg_solvers


class EmpiricSolution:
    """
    Unit helps design experiment via empiric distribution.
    """

    def __init__(self, power_calulation: Callable, desired_power: float, variating_param_name: List[str]) -> None:
        """
        power_calculation(**kwargs) - Empiric estiation of power of criterion
        """
        self.power_calc = power_calulation
        self.desired_power = desired_power
        self.var_param_name = variating_param_name

    def power(self, **kwargs) -> float:
        return self.power_calc(**kwargs)


class EmpiricSizeSolution(EmpiricSolution):
    """
    Special unit designing sample size via bootstrap.
    """

    def calc_upper_bound(self, **kwargs) -> int:
        deg: int = pkg_solvers.helper_bin_search_upper_bound_size(
            self.power_calc, self.desired_power, self.var_param_name, **kwargs
        )
        return 2**deg

    def calc_binary(self, **kwargs) -> int:
        upper_bound: int = self.calc_upper_bound(**kwargs) * 2
        return pkg_solvers.helper_binary_search_optimal_size(
            self.power_calc, self.desired_power, upper_bound, self.var_param_name, **kwargs
        )


class EmpiricEffectSolution(EmpiricSolution):
    """
    Special unit designing MDE via bootstrap.
    """

    def calc_upper_bound(self, **kwargs) -> float:
        deg: float = pkg_solvers.helper_bin_searh_upper_bound_effect(self.power_calc, self.desired_power, **kwargs)
        return 2**deg

    def calc_binary(self, **kwargs) -> int:
        upper_bound: float = self.calc_upper_bound(**kwargs)
        return pkg_solvers.helper_binary_search_effect_with_injection(
            self.power_calc, self.desired_power, upper_bound, self.var_param_name, **kwargs
        )
